'''Testing for utilities.py
'''

from mock import patch
import numpy as np
import numpy.testing as npt
import os
import pdb
import shutil
import subprocess
import unittest

import galaxy_dive.utils.utilities as utilities

########################################################################
########################################################################

class TestSmartDictStartup( unittest.TestCase ):

    def test_default( self ):

        d = { 'a' : 1, 'b' : 2 }

        smart_dict = utilities.SmartDict( d )

        self.assertEqual( smart_dict['b'], 2 )
        self.assertEqual( len( smart_dict ), 2 )

    def test_from_defaults_and_variations( self ):

        class TestClassA( object ):
            def __init__( self, a, b ):
                self.a = a
                self.b = b
            def return_contents( self ):
                return self.a, self.b

        defaults = { 'a' : 1, 'b' : 1 }
        variations = {
            1 : {},
            2 : { 'b' : 2 },
        }

        result = utilities.SmartDict.from_class_and_args( TestClassA, variations, defaults, )

        assert isinstance( result, utilities.SmartDict )

        expected = { 1 : ( 1, 1, ), 2 : ( 1, 2, ), }
        actual = result.return_contents()
        assert expected == actual

########################################################################

class TestSmartDict( unittest.TestCase ):

    def test_nested( self ):

        class TestClassA( object ):
            def __init__( self, key ):
                self.foo = 1234
                self.key = key
        class TestClassB( object ):
            def __init__( self, key ):
                self.test_class_a = TestClassA( key )
                self.key = key

        d = {}
        expected = {}
        expected2 = {}
        for i in range( 3 ):
            d[i] = TestClassB( i )
            expected[i] = 1234
            expected2[i] = i

        smart_d = utilities.SmartDict( d )

        actual = smart_d.test_class_a.foo
        self.assertEqual( expected, actual )

        actual = smart_d.key
        self.assertEqual( expected2, actual )

        actual = smart_d.test_class_a.key
        self.assertEqual( expected2, actual )

    ########################################################################

    def test_nested_method( self ):

        class TestClassA( object ):
            def foo( self, x ):
                return x**2
        class TestClassB( object ):
            def __init__( self ):
                self.test_class_a = TestClassA()

        d = {}
        expected = {}
        for i in range( 3 ):
            d[i] = TestClassB()
            expected[i] = 4

        smart_d = utilities.SmartDict( d )

        actual = smart_d.test_class_a.foo( 2 )

        self.assertEqual( expected, actual )

    ########################################################################

    def test_call_custom_kwargs( self ):

        class TestClassA( object ):
            def foo( self, x ):
                return x**2

        d = utilities.SmartDict( { 1 : TestClassA(), 2 : TestClassA(), } )

        kwargs = { 1 : { 'x' : 10}, 2 : { 'x' : 100}, }
        actual = d.foo.call_custom_kwargs( kwargs )
        expected = { 1 : 100, 2 : 10000, }
        self.assertEqual( expected, actual )

    ########################################################################

    def test_call_iteratively( self ):

        class TestClassA( object ):
            def foo( self, x ):
                return x**2

        d = utilities.SmartDict( { 1 : TestClassA(), 2 : TestClassA(), } )

        actual = d.foo.call_iteratively( [ 1, 2, ] )
        expected = { 1 : [1, 4,], 2 : [1, 4, ] }
        self.assertEqual( expected, actual )

    ########################################################################

    def test_multiply( self ):

        d = utilities.SmartDict( { 1 : 1, 2 : 2 } )

        expected = { 1 : 2, 2 : 4, }

        actual = d*2
        self.assertEqual( expected, actual )

        actual = 2*d
        self.assertEqual( expected, actual )

    ########################################################################

    def test_multiply_smart_dict( self ):

        d1 = utilities.SmartDict( { 1 : 1, 2 : 2 } )
        d2 = utilities.SmartDict( { 1 : 2, 2 : 3 } )

        expected = { 1 : 2, 2 : 6, }

        actual = d1*d2
        self.assertEqual( expected, actual )

        actual = d2*d1
        self.assertEqual( expected, actual )

    ########################################################################

    def test_divide( self ):

        d = utilities.SmartDict( { 1 : 2, 2 : 4 } )
        expected = { 1 : 1, 2 : 2, }
        actual = d/2
        self.assertEqual( expected, actual )

        d = utilities.SmartDict( { 1 : 2, 2 : 4 } )
        expected = { 1 : 2, 2 : 1, }
        actual = 4/d
        self.assertEqual( expected, actual )

    ########################################################################

    def test_divide_smart_dict( self ):

        d1 = utilities.SmartDict( { 1 : 9, 2 : 4 } )
        d2 = utilities.SmartDict( { 1 : 3, 2 : 2 } )

        expected = { 1 : 3, 2 : 2, }

        actual = d1/d2
        self.assertEqual( expected, actual )

    ########################################################################

    def test_add( self ):

        d = utilities.SmartDict( { 1 : 1, 2 : 2 } )
        expected = { 1 : 2, 2 : 3, }
        actual = d + 1
        self.assertEqual( expected, actual )

        d = utilities.SmartDict( { 1 : 1, 2 : 2 } )
        expected = { 1 : 2, 2 : 3, }
        actual = 1 + d
        self.assertEqual( expected, actual )

    ########################################################################

    def test_add_smart_dict( self ):

        d1 = utilities.SmartDict( { 1 : 9, 2 : 4 } )
        d2 = utilities.SmartDict( { 1 : 3, 2 : 2 } )

        expected = { 1 : 12, 2 : 6, }

        actual = d1 + d2
        self.assertEqual( expected, actual )

    ########################################################################

    def test_subtract( self ):

        d = utilities.SmartDict( { 1 : 1, 2 : 2 } )
        expected = { 1 : 0, 2 : 1, }
        actual = d - 1
        self.assertEqual( expected, actual )

        d = utilities.SmartDict( { 1 : 1, 2 : 2 } )
        expected = { 1 : 0, 2 : -1, }
        actual = 1 - d
        self.assertEqual( expected, actual )

    ########################################################################

    def test_subtract_smart_dict( self ):

        d1 = utilities.SmartDict( { 1 : 9, 2 : 4 } )
        d2 = utilities.SmartDict( { 1 : 3, 2 : 2 } )

        expected = { 1 : 6, 2 : 2, }

        actual = d1 - d2
        self.assertEqual( expected, actual )

    ########################################################################

    def test_sum_contents( self ):

        d1 = utilities.SmartDict( { 1 : 1, 2 : 2, 3 : 3, } )

        self.assertEqual( 6, d1.sum_contents() )

    ########################################################################

    def test_keymax_and_keymin( self ):

        d1 = utilities.SmartDict( { 1 : 1, 2 : 2, 3 : 3, } )

        self.assertEqual( (3, 3), d1.keymax() )
        self.assertEqual( (1, 1), d1.keymin() )

        d1 = utilities.SmartDict( { 5 : 4, 2 : -1, 3 : 0, } )

        self.assertEqual( (5, 4), d1.keymax() )
        self.assertEqual( (2, -1), d1.keymin() )

    ########################################################################

    def test_transpose( self ):

        d = utilities.SmartDict( {
            'a': { 1: 1, 2: 2, },
            'b': { 1: 3, 2: 4, },
        } )

        expected = utilities.SmartDict( {
            1: { 'a': 1, 'b': 3, },
            2: { 'a': 2, 'b': 4, },
        } )

        self.assertEqual( d.transpose(), expected )

    ########################################################################

    def test_array( self ):

        d = utilities.SmartDict( {
            'a' : 1,
            'c' : 2,
            'b' : 3,
        } )

        expected = np.array([ 1., 3., 2. ])

        actual = d.array()

        npt.assert_allclose( expected, actual )

########################################################################
########################################################################

class TestDictFromDefaultsAndVariations( unittest.TestCase ):

    def test_default( self ):

        defaults = { 'best cat' : 'Melvulu', }
        variations = {
            'person a' : { 'other best cat' : 'Chellbrat', },
            'person b' : {
                'best cat' : 'A Normal Melville Cat',
                'other best cat' : 'Chellcat',
            },
        }

        actual = utilities.dict_from_defaults_and_variations( defaults, variations )
        expected = {
            'person a' : {
                'best cat' : 'Melvulu',
                'other best cat' : 'Chellbrat',
            },
            'person b' : {
                'best cat' : 'A Normal Melville Cat',
                'other best cat' : 'Chellcat',
            },
        }

        for key in expected.keys():
            assert expected[key] == actual[key]

########################################################################
########################################################################

class TestGetCodeVersion( unittest.TestCase ):

    def test_default( self ):

        result = utilities.get_code_version( self )

        assert result is not None

    ########################################################################

    def test_works_in_otherdir( self ):

        cwd = os.getcwd()

        # Change directories
        os.chdir( '../' )

        result = utilities.get_code_version( self )

        assert result is not None

        # Change back
        os.chdir( cwd )

    ########################################################################

    def test_works_for_modules( self ):

        actual = utilities.get_code_version( utilities, instance_type='module' )

        expected = subprocess.check_output( [ 'git', 'describe', '--always'] )

        assert actual == expected

########################################################################
########################################################################

class TestGetInstanceSourceDir( unittest.TestCase ):

    def test_default( self ):

        actual = utilities.get_instance_source_dir( self )

        expected = os.path.join( os.getcwd(), 'tests/test_utils' )

        assert actual == expected

    ########################################################################

    def test_works_for_modules( self ):

        actual = utilities.get_instance_source_dir( utilities, 'module' )

        expected = os.path.join( os.getcwd(), 'utils' )

        assert actual == expected

########################################################################
########################################################################

class TestChunkList( unittest.TestCase ):

    def test_default( self ):

        n = 2
        l = range( 10 )
        expected = ( [ 0, 1, 2, 3, 4 ], [ 5, 6, 7, 8, 9 ] )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

        n = 3
        l = range( 10 )
        expected = ( [ 0, 1, 2, 3 ], [ 4, 5, 6 ], [ 7, 8, 9 ] )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

        n = 3
        l = range( 11 )
        expected = ( [ 0, 1, 2, 3 ], [ 4, 5, 6, 7 ], [ 8, 9, 10 ] )
        actual = utilities.chunk_list( l, n )
        for expected_, actual_ in zip( expected, actual ):
            assert expected_ == actual_

########################################################################
########################################################################

class TestArraySetConversion( unittest.TestCase ):

    def test_arrays_to_set( self ):

        expected = set( [ ( 1, 1 ), ( 2, 2 ), ( 0, 0 ) ] )

        actual = utilities.arrays_to_set( np.arange( 3 ), np.arange( 3 ) )

        self.assertEqual( expected, actual )

    ########################################################################

    def test_arrays_to_set_long( self ):

        expected = set( [ ( 1, 1, 1 ), ( 2, 2, 2 ), ( 0, 0, 0 ) ] )

        actual = utilities.arrays_to_set( np.arange( 3 ), np.arange( 3 ), np.arange(3) )

        self.assertEqual( expected, actual )

    ########################################################################

    def test_set_to_arrays( self ):

        expected = np.array([
            [ 1, 4, 7, ],
            [ 2, 5, 8, ],
            [ 3, 6, 9, ],
        ])

        actual = utilities.set_to_arrays( set( [ ( 1, 2, 3, ), ( 4, 5, 6 ), ( 7, 8, 9 ), ] ) )

        npt.assert_allclose( expected[:,0], actual[:,-1] )

        # To allow for full comparison. We don't really care what order they're in, as long as they're matched.
        actual.sort()

        npt.assert_allclose( expected, actual )

########################################################################
########################################################################

class TestStoreParameters( unittest.TestCase ):

    def test_basic( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, a, b, c=3 ):
                pass

        o = Object( 1, 2, )

        self.assertEqual( 1, o.a )
        self.assertEqual( 2, o.b )
        self.assertEqual( 3, o.c )
        self.assertEqual( [ 'a', 'b', 'c' ], o.stored_parameters )

    ########################################################################

    def test_kwargs_input( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, a, b, c=3, **kwargs ):
                pass

        o = Object( **{ 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 } )

        self.assertEqual( 1, o.a )
        self.assertEqual( 2, o.b )
        self.assertEqual( 4, o.c )
        self.assertEqual( {'d' : 5}, o.kwargs )
        self.assertEqual( [ 'a', 'b', 'c', 'kwargs', ], o.stored_parameters )

    ########################################################################

    def test_args_and_kwargs_input( self ):

        class Object( object ):

            @utilities.store_parameters
            def __init__( self, *args, **kwargs ):
                pass

        args = ( 'dog', )
        kwargs = { 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 }
        o = Object( *args, **kwargs )

        self.assertEqual( ('dog',), o.args )
        self.assertEqual( { 'a' : 1, 'b' : 2, 'c' : 4, 'd' : 5 }, o.kwargs )
        self.assertEqual( [ 'args', 'kwargs', ], o.stored_parameters )

########################################################################
########################################################################

class TestGenerateDerivedDataDoc( unittest.TestCase ):

    def setUp( self ):

        self.output_filepath = './tests/data/derived_data.rst'

        if os.path.isfile( self.output_filepath ):
            os.remove( self.output_filepath )

    ########################################################################

    def test_gen_derived_data_doc( self ):

        object_import_path = 'galaxy_dive.tests.test_utils.test_utilities.GenDerivedDataObject'

        # Call the function
        utilities.gen_derived_data_doc(
            object_import_path,
            self.output_filepath,
        )
                
        # Build Expected
        # Get the header
        with open ( './utils/derived_data_header.rst', 'r') as myfile:
            expected_header = myfile.readlines()
        # Get the generated data
        expected_generated = [
            '\n',
            '.. autoclass:: {}\n'.format( object_import_path ),
            '    :show-inheritance:\n',
            '\n',
            '    .. automethod:: calc_A\n',
            '    .. automethod:: calc_B\n',
        ]
        expected = expected_header + expected_generated

        with open ( self.output_filepath, 'r') as myfile:
            actual = myfile.readlines()

        self.assertEqual( expected, actual )

# Test Object for TestGenerateDerivedDataDoc
class GenDerivedDataObject( object ):

    def method_a( self, *args, **kwargs ):
        '''Method a.'''
        pass

    def calc_A( self ):
        '''Calculate A''' 
        pass

    def calc_B( self ):
        '''Calculate A''' 
        pass



#!/usr/bin/env python
'''Subclass for analyzing particle data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import h5py
import numpy as np

import galaxy_dive.utils.io as io
import galaxy_dive.analyze_data.simulation_data as simulation_data
import galaxy_dive.utils.utilities as utilities

########################################################################

class GriddedData( simulation_data.SnapshotData ):
  '''Class for handling data that forms a Cartesian grid.'''

  @utilities.store_parameters
  def __init__( self, sdir=None, snum=None, Nx=None, gridsize=None, ionized=None, ion_grid=False, **kwargs ):
    '''
    Args:
      sdir (str) : Directory the snaphost is stored in
      snum (str) : Snapshot number
      Nx (int) : Number of grid cells on a side
      gridsize (float or str) : How large the grid is
      ionized (str) : Legacy name. Originally how the ionization state is calculated. 
                        Currently any identifying string added to the end of the filename.
                        'R13' : Ionization state calculated using Rahmati+13 fitting function
      ion_grid (bool) : Whether or not this is a grid containing ion information.
    '''

    # Note that we assume the grid is centered.
    super( GriddedData, self ).__init__( data_dir=sdir, snum=snum, centered=True, **kwargs )

    self.retrieve_data()

  ########################################################################

  def retrieve_data( self ):

    # Open file
    self.grid_file_name = io.getGridFilename( self.sdir, self.snum, self.Nx, self.gridsize, self.ionized, \
                          ion_grid=self.ion_grid )
    f = h5py.File( self.grid_file_name, 'r' )

    # Get the grid attributes
    self.data_attrs = {}
    for key in f.attrs.keys():
      self.data_attrs[key] = f.attrs[key]

    def load_data(f):

      # Get the data
      self.data = {}
      for key in f.keys():

        # Parse the keys and put in a more general format
        # All units should be the standard GIZMO output units
        # Gas Density
        if key == 'GasDen':
          self.data['Den'] = f[key][...]
        # Gas Neutral Hydrogen fraction
        elif key == 'GasNHI':
          self.data['nHI'] = f[key][...]
        # Should be the mean free electron number per proton
        elif key == 'GasNe':
          self.data['ne'] = f[key][...]
        # Same as above, but due to helium
        elif key == 'GasNe_He':
          self.data['ne_He'] = f[key][...]
        # Star Formation Rate
        elif key == 'GasSFR':
          self.data['SFR'] = f[key][...]
        # Temperature
        elif key == 'GasT':
          self.data['T'] = f[key][...]
        # X Velocity
        elif key == 'GasVx':
          vx = f[key][...]
        # Y Velocity
        elif key == 'GasVy':
          vy = f[key][...]
        # Z Velocity
        elif key == 'GasVz':
          vz = f[key][...]
        # Metallicity
        elif key == 'GasZ':
          self.data['Z'] = f[key][...]
        else:
          raise Exception('NULL key')

      # Finish organizing velocity
      self.data['V'] = np.array([vx, vy, vz])

    def load_ion_data(f):

      self.data = {}
      for key in f.keys():
        self.data[key] = f[key][...]

    # Load the data
    if not self.ion_grid:
      load_data(f)
    elif self.ion_grid:
      load_ion_data(f)
    else:
      raise Exception('Unrecognized ion_grid')

    f.close()

  ########################################################################
  ########################################################################

  def handle_data_key_error( self, data_key ):
    '''When get_data() fails to data_key in self.data, it passes the data_key to try and generate that data.

    Args:
      data_key (str) : Key to try and generate data for

    Modifies:
      self.data[data_key] (np.array) : If it finds a function to generate the data, it will do so
    '''

    if data_key ==  'Rx' or data_key ==  'Ry' or data_key ==  'Rz' or data_key == 'P':
      self.calc_positions()
    elif data_key[:-3] ==  'Rx_face' or data_key[:-3] ==  'Ry_face' or data_key[:-3] ==  'Rz_face':
      self.calc_face_positions( data_key )
    elif data_key[:-3] ==  'R_face':
      self.calc_impact_parameter( data_key )
    elif data_key == 'M':
      self.calc_mass()

    else:
      super( GriddedData, self ).handle_data_key_error( data_key )

  ########################################################################
  ########################################################################

  def calc_positions( self ):
    '''Calculate the positions of gridcells.'''

    # Get how the spacing on one side
    Nx = self.data.values()[0].shape[0]
    gridsize = self.data_attrs['gridsize']
    pos_coords = np.linspace(-gridsize/2., gridsize/2., Nx)

    # Mesh the spacing together for all the grid cells
    x, y, z = np.meshgrid(pos_coords, pos_coords, pos_coords, indexing='ij')

    self.data['P'] = np.array([x, y, z])

  ########################################################################

  def calc_face_positions( self, data_key ):
    '''Calculate positions if you're just looking at one face of a grid.'''

    # Figure out which face to calculate for
    target_face = data_key.split( '_' )[-1]

    if target_face == 'xy':
      self.data['Rx_face_xy'] = self.get_data( 'Rx' )[:, :, 0]
      self.data['Ry_face_xy'] = self.get_data( 'Ry' )[:, :, 0]
    elif target_face == 'xz':
      self.data['Rx_face_xz'] = self.get_data( 'Rx' )[:, 0, :]
      self.data['Rz_face_xz'] = self.get_data( 'Rz' )[:, 0, :]
    elif target_face == 'yz':
      self.data['Ry_face_yz'] = self.get_data( 'Ry' )[0, :, :]
      self.data['Rz_face_yz'] = self.get_data( 'Rz' )[0, :, :]

  ########################################################################

  def calc_impact_parameter( self, data_key ):

    # Figure out which face to calculate for
    target_face = data_key.split( '_' )[-1]
    
    if target_face == 'xy':
      self.data[data_key] = np.sqrt( self.get_data( 'Rx_face_xy' )**2. + self.get_data( 'Ry_face_xy' )**2. ) 
    elif target_face == 'xz':
      self.data[data_key] = np.sqrt( self.get_data( 'Rx_face_xz' )**2. + self.get_data( 'Ry_face_xz' )**2. ) 
    elif target_face == 'yz':
      self.data[data_key] = np.sqrt( self.get_data( 'Rx_face_yz' )**2. + self.get_data( 'Ry_face_yz' )**2. ) 

  ########################################################################

  def calc_mass( self ):
    '''Calculate the mass per grid cell.'''

    # Calculate the mass per grid cell, for easy use later
    self.data['M'] = self.data['Den']*(self.data_attrs['gridsize']/self.data_attrs['Nx'])**3.

  ########################################################################

  def calc_column_density( self, key, face ):
    '''Calculate the column density for a targeted key.

    Args --
    key (str) : The data to project.
    face (int) : The face to use. 0=yz, 1=xz, 2=xy.

    Returns --
    key_col_den (np.array) : The column density in {key units}/cm^2. Assumes the grid uses fiducial GIZMO units.
    '''

    key_data = self.get_data( key )

    # Get the projected data
    summed_key_data = key_data.sum(face)

    # Get the column density
    self.data_attrs['cell_length'] = self.data_attrs['gridsize']/float( self.data_attrs['Nx'] - 1 )
    dx = self.data_attrs['cell_length']*3.086e21
    key_col_den = dx*summed_key_data

    return key_col_den





'''Testing for data_operations.py
'''

from mock import patch
import numpy as np
import numpy.testing as npt
import pdb
import unittest

import galaxy_dive.utils.data_operations as data_operations

########################################################################

class SaneEqualityArray(np.ndarray):
    '''Numpy array subclass that allows you to test if two arrays are equal.'''

    def __eq__(self, other):
            return (isinstance(other, np.ndarray) and self.shape == other.shape and np.allclose(self, other))

def sane_eq_array(list_in):
    '''Wrapper for SaneEqualityArray, that takes in a list.'''

    arr = np.array(list_in)

    return arr.view(SaneEqualityArray)

########################################################################

class TestRotateCoordinates(unittest.TestCase):

    def setUp(self):

        # What is the name of the function.
        self.fn = data_operations.rotate_coordinates

        self.input_arr = np.random.rand(4, 3)
        self.angle = -np.pi/2.
        self.axis = np.array([0., 0., 1.])

        self.default_args = [self.input_arr, self.angle, self.axis]

    ########################################################################

    def test_output_format(self):
        '''Test that we return an array of the same dimensions'''

        expected = self.input_arr.shape

        result = self.fn(*self.default_args)

        actual = result.shape

        self.assertEqual(expected, actual)

    ########################################################################

    def test_rotate_vector_in_xy_plane_by_90(self):

        # Change the input array.
        self.default_args[0] = np.array([1., 0., 0.])

        result = self.fn(*self.default_args)

        expected = np.array([0., 1., 0.])

        npt.assert_allclose(result, expected, atol=1e-8)

        # Second test
        # Change the input array.
        self.default_args[0] = np.array([0., 1., 0.])

        result = self.fn(*self.default_args)

        expected = np.array([-1., 0., 0.])

        npt.assert_allclose(result, expected, atol=1e-8)

        # Third test
        # Change the input array.
        self.default_args[0] = np.array([0., 0., 1.])

        result = self.fn(*self.default_args)

        expected = np.array([0., 0., 1.])

        npt.assert_allclose(result, expected, atol=1e-8)

    ########################################################################

    def test_rotate_array_in_xy_plane_by_90(self):

        # Change the input array.
        self.default_args[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.], [0., 1., 0.]])

        result = self.fn(*self.default_args)

        expected = np.array([[0., 1., 0.], [-1., 0., 0.], [0., 0., 1.], [-1., 0., 0.],])

        npt.assert_allclose(expected, result)

    ########################################################################

    def test_rotate_array_down_x_axis_by_90(self):

        # Change the input array.
        self.default_args[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.], [0., 1., 0.]])

        # Change the axis
        self.default_args[2] = np.array([1., 0., 0.])

        result = self.fn(*self.default_args)

        expected = np.array([[1., 0., 0.], [0., 0., 1.], [0., -1., 0.], [0., 0., 1.],])

        npt.assert_allclose(expected, result)

    ########################################################################

    def test_rotate_by_arbitrary_amount(self):

        # Change the input array.
        self.default_args[0] = np.array([[1., 0., 0.], [0., 1., 0.], [2., 0., 0.]])

        # Change the angle to an arbitrary angle
        self.default_args[1] = np.pi/5.

        # Change the vector to an arbitrary vector
        self.default_args[2] = np.array([1., 2., 3. ])

        result = self.fn(*self.default_args)

        # Test we keep the same angle with the vector we're rotating around
        expected_dot_w_rot_vector = np.dot(self.default_args[0][0], self.default_args[2])
        actual_dot_w_rot_vector = np.dot(result[0], self.default_args[2])
        npt.assert_allclose(expected_dot_w_rot_vector, actual_dot_w_rot_vector)

        # Test we get out the magnitude we're expecting
        expected_magnitude = 2.
        actual_magnitude = np.linalg.norm(result[2])
        npt.assert_allclose(expected_magnitude, actual_magnitude)

        # Test we get out the rotation we're expecting
        # (In the sense of the dot product should be np.pi/2, i.e. both should be rotated the same amount)
        expected_rotation = np.cos(-np.pi/2.)
        actual_rotation = np.dot(result[1], result[0])
        npt.assert_allclose(expected_rotation, actual_rotation, atol=1e-7)

        # Test we get out the rotation we're expecting
        # (In the sense of the dot product of the projection should be the angle it's rotated by.)
        expected_rotation = np.cos(self.default_args[1])
        dir_unit = self.default_args[2]/np.linalg.norm(self.default_args[2])
        # Test for each part of the coordinates
        for result_, orig_ in zip(result, self.default_args[0]):

            # Get the projections
            orig_proj = orig_ - np.dot(dir_unit, orig_)*dir_unit
            result_proj = result_ - np.dot(dir_unit, result_)*dir_unit

            actual_rotation = np.dot(orig_proj, result_proj)/np.linalg.norm(orig_proj)/np.linalg.norm(result_proj)

            npt.assert_allclose(expected_rotation, actual_rotation, atol=1e-14)

########################################################################

class TestAlignAxes(unittest.TestCase):

    def setUp(self):

        # What is the name of the function.
        self.fn = data_operations.align_axes

        self.input_arr = np.random.rand(4, 3)
        self.new_z_axis = np.array([1., 2., 3.])

        self.default_args = [self.input_arr, self.new_z_axis]

    ########################################################################

    def test_output_format(self):
        '''Test that we return an array of the same dimensions'''

        expected = self.input_arr.shape

        result = self.fn(*self.default_args)

        actual = result.shape

        self.assertEqual(expected, actual)

    ########################################################################

    def test_rotate_array_to_align_z_with_x(self):

        # Change the input array.
        self.default_args[0] = np.array([ 1., 0., 0., ])
        #self.default_args[0] = np.array([0., 1., 0.])

        # Align z with x
        self.default_args[1] = np.array([ 1., 0., 0. ])
        result = self.fn(*self.default_args)

        #expected = np.array([0., 0., 1.])
        expected = np.array([0., 0., 1.])

        npt.assert_allclose(expected, result)

    ########################################################################

    def test_rotate_array_to_align_z_with_x_and_check_products(self):

        # Change the input array.
        self.default_args[0] = np.array([
            [1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.],
            [1., 0., 0.],
        ])

        # Align z with x
        self.default_args[1] = np.array([1., 0., 0.])

        result = self.fn(*self.default_args)

        # Make sure that our results have the right angles
        expected = 0.
        actual = np.dot(self.default_args[0][0], result[0])
        npt.assert_allclose(expected, actual, atol=1e-14)

        # Make sure that our results have the right angles
        expected = 1.
        actual = np.dot(np.array([0., 0., 1.]), result[0])
        npt.assert_allclose(expected, actual, atol=1e-14)

        # Make sure that our results have the right angles
        expected = 0.
        actual = np.dot(np.array([0., 0., 1.]), result[1])
        npt.assert_allclose(expected, actual, atol=1e-14)

        # Make sure that our results have the right angles
        expected = 0.
        actual = np.dot(np.array([0., 0., 1.]), result[2])
        npt.assert_allclose(expected, actual, atol=1e-14)

    ########################################################################

    def test_rotate_doesnt_do_anything(self):
        '''Test that nothing happens when new_z = old_z'''

        # Change the input array.
        self.default_args[0] = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.], [1., 0., 0.]])

        # Align z with z
        self.default_args[1] = np.array([0., 0., 1.])

        # Add an arbitrary point
        self.default_args.append( np.array([1., 2., 3.]) )

        result = self.fn(*self.default_args)

        npt.assert_allclose(self.default_args[0], result)

    ########################################################################

    def test_rotate_about_point(self):

        # Change the input array.
        self.default_args[0] = np.array([1., 0., 0.],)

        # Align z with y
        self.default_args[1] = np.array([0., 1., 0.])

        # Add an arbitrary point
        self.default_args.append( np.array([0., 0., 1.]) )

        result = self.fn(*self.default_args)

        # Test if we have the transformed arrays right
        npt.assert_allclose( result, np.array([1., 1., 1.]) )

    ########################################################################

    def test_rotate_about_point_arr(self):
        '''Test that we can do an offset for a number of coordinates.'''

        # Change the input array.
        self.default_args[0] = np.array([[1., 0., 0.], [0., 1., 1.], [0., 0., 1.], [1., 0., 0.]])

        # Align z with y
        self.default_args[1] = np.array([0., 1., 0.])

        # Add an arbitrary point
        self.default_args.append( np.array([0., 0., 1.]) )

        result = self.fn(*self.default_args)

        # Test if we have the transformed arrays right
        npt.assert_allclose( result[0], np.array([1., 1., 1.]) )
        npt.assert_allclose( result[1], np.array([0., 0., 2.]) )

    ########################################################################

    def test_align_vectors( self ):
        '''Test that we can align with a random vector.'''

        # Vector to align
        v = np.random.uniform( -1., 1., 3 )

        # Build the right answer
        v_norm = np.linalg.norm( v )
        expected = np.tile( v/v_norm, (3, 1) )

        # Create the input
        self.default_args[0] = np.array([
            [ 0., 0., 1., ],
            [ 0., 0., 1., ],
            [ 0., 0., 1., ],
        ])

        # Align z with vector
        self.default_args[1] = v

        # Actual calculation
        actual = self.fn( align_frame=False, *self.default_args )

        npt.assert_allclose( expected, actual )

########################################################################
########################################################################

class TestCumsum2D( unittest.TestCase ):

    def test_cumsum2d( self ):

        arr = np.array([
              [6, 5, 1],
              [7, 0, 1],
              [1, 1, 8]]
        )

        expected = np.array([
              [6, 11, 12],
              [13, 18, 20],
              [14, 20, 30]]
        )

        actual = data_operations.cumsum2d( arr )

        npt.assert_allclose( expected, actual )

    ########################################################################

    def test_cumsum2d_dir( self ):

        arr = np.array([
              [6, 5, 1],
              [7, 0, 1],
              [1, 1, 8],
        ])

        expected = np.array([
              [12, 6, 1],
              [20, 7, 2],
              [30, 16, 10],
        ])

        actual = data_operations.cumsum2d( arr, directions=[1,-1] )

        npt.assert_allclose( expected, actual )
                



#!/usr/bin/env python
'''General astronomy utilities

@author: Daniel Angles-Alcazar, Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import numpy as np

import galaxy_dive.utils.constants as constants

########################################################################

def hubble_parameter( redshift, h=0.702, omega_matter=0.272, omega_lambda=0.728, units='km/s/Mpc' ):
  '''Return Hubble factor in 1/sec for a given redshift.

  Args:
    redshift (float): The input redshift.
    h (float): The hubble parameter.
    omega_matter (float): Cosmological mass fraction of matter.
    omega_lambda (float): Cosmological mass fraction of dark energy.
    units (str): The units the hubble parameter should be returned in.

  Returns:
    hubble_a (float): Hubble factor in specified units
  '''

  zp1 = (1. + redshift)

  e_z = np.sqrt( omega_matter*zp1**3. + omega_lambda )

  hubble_z = h*100.*e_z

  if units == 'km/s/Mpc':
    pass
  elif units == 'km/s/kpc':
    hubble_z /= 1e3
  elif units == '1/s':
    hubble_z /= constants.KM_PER_KPC*1e3
  else:
    raise KeyError( "Unspecified units, {}".format( units ) )
    

  return hubble_z

########################################################################

def age_of_universe( redshift, h=0.71, omega_matter=0.27 ):
  '''Get the exact solution to the age of universe (for a flat universe) to a given redshift

  Args:
    redshift (float): The input redshift.
    h (float): The hubble parameter.
    omega_matter (float): Cosmological mass fraction of matter.

  Returns:
    t (float): Age of the universe in Gyr
  '''

  a = 1./(1.+redshift)
  x = omega_matter / (1. - omega_matter) / (a*a*a)

  t = (2./(3.*np.sqrt(1. - omega_matter))) * np.log( np.sqrt(x) / (-1. + np.sqrt(1.+x)) )

  t *= 13.777 * (0.71/h) ## in Gyr

  return t

########################################################################

def circular_velocity( r_vir, m_vir ):
  '''Calculate the circular velocity of a halo in km/s.

  Args:
    r_vir (float or array-like) : The virial radius in pkpc.
    m_vir (float or array-like) : The halo mass in Msun.

  Returns:
    v_c : Circular velocity of the halo in km/s, indexed the same way that ahf_reader.mtree_halos[i] is.
  '''
  
  # Convert r_vir and m_vir to cgs
  r_vir_cgs = r_vir*constants.CM_PER_KPC
  m_vir_cgs = m_vir*constants.MSUN

  # Get v_c
  v_c_cgs = np.sqrt( constants.G_UNIV * m_vir_cgs / r_vir_cgs )

  # Convert to km/s
  v_c = v_c_cgs / constants.CM_PER_KM

  return v_c

########################################################################

def get_sneii_metal_budget( m_star, y_z_ii=0.030 ):
    '''Get the total mass of metal produced by SNe by z=0 in a galaxy with a
    z=0 stellar mass of m_star.
    This is taken from Peeple+2014, who use the star formation histories of
    Leitner+2012.

    Args:
        m_star (float or array-like) :
            Mass of the target galaxy at z=0 in units of Msun.

        y_z_ii (float) :
            Nucleosynthetic yield of all heavy 

    Returns:
        sneii_metal_budget (same as m_star):
            Units of Msun.
    '''

    return 1.2856 * y_z_ii * m_star ** 1.0146



#!/usr/bin/env python
'''Class for plotting simulation data.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

# Base python imports
import numpy as np
import os
import scipy.signal as signal
import warnings

import matplotlib
matplotlib.use('PDF')
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as plt_colors
import matplotlib.gridspec as gridspec
import matplotlib.patheffects as path_effects
import matplotlib.transforms as transforms

import galaxy_dive.utils.mp_utils as mp_utils
import galaxy_dive.utils.utilities as utilities

import plotting as gen_plot
import pu_colormaps as pu_cm

########################################################################

default = object()

########################################################################
########################################################################

class GenericPlotter( object ):

    @utilities.store_parameters
    def __init__( self, data_object, label=None, color='black', ):
        '''
        Args:
            data_object ( generic_data.GenericData object or subclass of such ) : The data container to use.
        '''

        pass

    ########################################################################
    # Alternate inherent methods
    ########################################################################

    def __getattr__( self, attr):
        '''By replacing getattr with the following code, we allow automatically searching the data_object
        for the appropriate attribute as well, while losing none of the original functionality.
        '''

        print( "Attribute {} not found in plotting object. Checking data object.".format( attr ) )

        return getattr( self.data_object, attr )

    ########################################################################
    # Specific Generic Plots
    ########################################################################

    def histogram( self,
        data_key,
        provided_data = None,
        provided_hist = None,
        weight_key = default,
        slices = None,
        ax = default,
        fix_invalid = False,
        mask_zeros = False,
        invalid_fix_method = default,
        bins = 32,
        normed = True,
        norm_type = 'probability',
        scaling = None,
        smooth = False,
        smoothing_window_length = 9,
        smoothing_polyorder = 3,
        histogram_style = 'step',
        color = 'black',
        linestyle = '-',
        linewidth = 3.5,
        alpha = 1.,
        x_range = default, y_range = default,
        x_label = default, y_label = default,
        add_x_label = True, add_y_label = True,
        plot_label = default,
        line_label = default,
        label_fontsize = 24,
        x_scale = 'linear', y_scale = 'linear',
        cdf = False,
        vertical_line = None,
        vertical_line_kwargs = { 'linestyle': '--', 'linewidth': 3, 'color': 'k', },
        return_dist = False,
        assert_contains_all_data = True,
        data_kwargs = {},
        *args, **kwargs ):
        '''Make a 2D histogram of the data. Extra arguments are passed to self.data_object.get_selected_data().

        Args:
            data_key (str) :
                Data key to plot.

            weight_key (str) :
                Data key for data to use as a weight. By default, no weight.

            slices (int or tuple of slices) :
                How to slices the data.

            ax (axis) :
                What axis to use. By default creates a figure and places the axis on it.

            fix_invalid (bool) :
                Throw away invalid values?

            invalid_fix_method (float or int) :
                How to handle invalid values. By default throw them away. Providing a value to this argument instead replaces
                them with that value.

            bins (int or array-like) :
                bins argument to be passed to np.histogram

            normed (bool) :
                Normalize the histogram?

            color (str) :
                Color of histogram.

            linestyle (str) :
                Linestyle of histogram.

            linewidth (float) :
                Linewidth for histogram.

            alpha (float) :
                Alpha value for histogram

            x_range, y_range ( [min,max] ) :
                What are the minimum and maximum x- and y- values to include?
                Defaults to matplotlib's automatic choices

            x_label, ylabel (str) :
                Axes labels. Defaults to the data_key for the x-axis and "Normalized Histogram" for the y-axis.

            add_x_label, add_y_label (bool) :
                Include axes labels?

            plot_label (str or dict) :
                What to label the plot with. By default, uses self.label.

            line_label (str) :
                What label to give the line.

            label_fontsize (int) :
                Fontsize for the labels.

            x_scale, y_scale (str) :
                What scales to use for the x and y axes.

            cdf (bool) :
                Plot a CDF instead.

            vertical_line (float) :
                Plot a vertical line at this value on the x-axis, if true.

            vertical_line_kwargs (dict) :
                Arguments to pass to ax.plot( for the vertical line.

            return_dist (bool) :
                If True, return the data values and the edges for the histogram.

            assert_contains_all_data (bool) :
                If True, make sure that the histogram plots all selected data.

            *args, **kwargs :
                Extra arguments to pass to self.data_object.get_selected_data()
        '''

        print( "Plotting histogram for {}".format( data_key ) )

        if provided_hist is None:

            if isinstance( slices, int ):
                sl = ( slice(None), slices )
            else:
                sl = slices

            data_kwargs = utilities.merge_two_dicts( data_kwargs, kwargs )

            if provided_data is None:
                data = self.data_object.get_selected_data(
                    data_key,
                    sl=sl,
                    *args, **data_kwargs
                ).copy()
            else:
                data = provided_data.copy()

            if weight_key is default:
                weights = None
            else:

                if 'scale_key' in kwargs:
                    warnings.warn(
                        "Scaling weights by {}. Is this correct?".format(
                            kwargs['scale_key']
                        )
                    )
                weights = self.data_object.get_selected_data( weight_key, sl=sl, *args, **kwargs )

            if fix_invalid:
                if invalid_fix_method is default:
                    data = np.ma.fix_invalid( data ).compressed()
                else:
                    data = np.ma.fix_invalid( data )
                    data.fill_value = invalid_fix_method
                    data = data.filled()

            # Make the histogram itself
            hist, edges = np.histogram( data, bins=bins, weights=weights )

            # Make sure we have all the data in the histogram
            if assert_contains_all_data:
                assert data.size == hist.sum()

            if normed:

                if norm_type == 'probability':
                    hist = hist.astype( float ) / ( hist.sum()*(edges[1] - edges[0]) )
                elif norm_type == 'bin_width':
                    hist = hist.astype( float ) / (edges[1] - edges[0])
                elif norm_type == 'outer_edge':
                    hist = hist.astype( float ) / hist[-1]
                elif norm_type == 'max_value':
                    hist = hist.astype( float ) / hist.max()
                else:
                    raise Exception(
                        "Unrecognized norm_type, {}".format( norm_type )
                    )

            if scaling is not None:
                hist *= scaling

            if cdf:
                hist = np.cumsum( hist )*(edges[1] - edges[0])

        else:
            hist = provided_hist
            edges = bins

        if mask_zeros:
            hist = np.ma.masked_where(
                hist < 1e-14,
                hist,
            )

        if smooth:
            hist = signal.savgol_filter(
                hist,
                window_length = smoothing_window_length,
                polyorder = smoothing_polyorder,
            )


        if ax is default:
            fig = plt.figure( figsize=(11,5), facecolor='white', )
            ax = plt.gca()
        if line_label is default:
            line_label = self.label

        if color is default:
            color = self.color

        # Inserting a 0 at the beginning allows plotting a numpy histogram with a step plot
        if histogram_style == 'step':
            ax.step(
                edges,
                np.insert(hist, 0, 0.),
                color=color,
                linestyle=linestyle,
                linewidth=linewidth,
                label=line_label,
                alpha=alpha,
            )
        elif histogram_style == 'line':
            x_values = 0.5 * ( edges[:-1] + edges[1:] )
            ax.plot(
                x_values,
                hist,
                color=color,
                linestyle=linestyle,
                linewidth=linewidth,
                label=line_label,
                alpha=alpha,
            )
        else:
            raise KeyError(
                "Unrecognized histogram_style, {}".format( histogram_style )
            )

        # Plot a vertical line?
        if vertical_line is not None:
            trans = transforms.blended_transform_factory( ax.transData, ax.transAxes )
            ax.plot(
                [ vertical_line, ]*2,
                [ 0., 1., ],
                transform = trans,
                **vertical_line_kwargs
)

        # Plot label
        if plot_label is default:
            plt_label = ax.annotate(
                s = self.label,
                xy = (0.,1.0),
                va = 'bottom',
                xycoords = 'axes fraction',
                fontsize = label_fontsize,
             )
        elif isinstance( plot_label, str ):
            plt_label = ax.annotate(
                s = plot_label,
                xy = (0.,1.0),
                va = 'bottom',
                xycoords = 'axes fraction',
                fontsize = label_fontsize,
             )
        elif isinstance( plot_label, dict ):
            plt_label = ax.annotate( **plot_label )
        elif plot_label is None:
            pass
        else:
            raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )

        # Add axis labels
        if add_x_label:
            if x_label is default:
                x_label = data_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is default:
                if not cdf:
                    y_label = r'Normalized Histogram'
                else:
                    y_label = r'CDF'
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        if x_range is not default:
            ax.set_xlim( x_range )
        if y_range is not default:
            ax.set_ylim( y_range )

        ax.set_xscale( x_scale )
        ax.set_yscale( y_scale )

        if return_dist:
            return hist, edges

    ########################################################################

    def histogram2d( self,
        x_key, y_key,
        weight_key = default,
        x_data_args = {}, y_data_args = {},
        weight_data_args = {},
        slices = None,
        ax = default,
        x_range = default, y_range = default,
        x_scale = 'linear', y_scale = 'linear', z_scale = 'log',
        n_bins = 128,
        average = False,
        normed = False,
        hist_div_arr = None,
        conditional_y = False,
        y_div_function = None,
        vmin = None, vmax = None,
        min_bin_value_displayed = None,
        zorder = 0,
        add_colorbar = True,
        cmap = pu_cm.magma,
        colorbar_args = default,
        x_label = default, y_label = default,
        add_x_label = True, add_y_label = True,
        plot_label = default,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = False,
        label_fontsize = 24,
        tick_param_args = default,
        save_file = default,
        out_dir = None,
        fix_invalid = True,
        line_slope = default,
        cdf = False,
        horizontal_line = None, vertical_line = None,
        horizontal_line_kwargs = { 'linestyle': '--', 'linewidth': 5, 'color': '#337DB8', },
        vertical_line_kwargs = { 'linestyle': '--', 'linewidth': 5, 'color': '#337DB8', },
        return_dist = False,
        *args, **kwargs ):
        '''Make a 2D histogram of the data. Extra arguments are passed to get_selected_data.
        Args:
            x_key, y_key (str) : Data keys to plot.
            weight_key (str) : Data key for data to use as a weight. By default, no weight.
            x_data_args, y_data_args (dicts) : Keyword arguments to be passed only to x or y.
            slices (int or tuple of slices) : How to slices the data.
            ax (axis) : What axis to use. By default creates a figure and places the axis on it.
            x_range, y_range ( (float, float) ) : Histogram edges. If default, all data is enclosed. If list, set manually.
                If float, is +- x_range*length scale at that snapshot.
            n_bins (int) : Number of bins in the histogram.
            vmin, vmax (float) : Limits for the colorbar.
            aspect (str) : What should the aspect ratio of the plot be?
            plot_halos (bool) : Whether or not to plot merger tree halos on top of the histogram.
                Only makes sense for when dealing with positions.
            add_colorbar (bool) : If True, add a colorbar to colorbar_args
            colorbar_args (axis) : What axis to add the colorbar to. By default, is ax.
            x_label, ylabel (str) : Axes labels.
            add_x_label, add_y_label (bool) : Include axes labels?
            plot_label (str or dict) : What to label the plot with. By default, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            tick_param_args (args) : Arguments to pass to ax.tick_params. By default, don't change inherent defaults.
            out_dir (str) : If given, where to save the file.
            fix_invalid (bool) : Fix invalid values.
            line_slope (float) : If given, draw a line with the given slope.
        '''

        if isinstance( slices, int ):
            sl = ( slice(None), slices )
        else:
            sl = slices

        varying_kwargs = {
            'x': x_data_args,
            'y': y_data_args,
            'weight': weight_data_args
        }
        data_kwargs = utilities.dict_from_defaults_and_variations( kwargs, varying_kwargs )

        # Get data
        x_data = self.data_object.get_selected_data( x_key, sl=sl, *args, **data_kwargs['x'] ).copy()
        y_data = self.data_object.get_selected_data( y_key, sl=sl, *args, **data_kwargs['y'] ).copy()

        if y_div_function is not None:
            y_div_values = y_div_function( x_data )
            y_data /= y_div_values

        # Fix NaNs
        if fix_invalid:
            x_mask = np.ma.fix_invalid( x_data ).mask
            y_mask = np.ma.fix_invalid( y_data ).mask
            mask = np.ma.mask_or( x_mask, y_mask )

            x_data = np.ma.masked_array( x_data, mask=mask ).compressed()
            y_data = np.ma.masked_array( y_data, mask=mask ).compressed()

        if weight_key is default:
            weights = None
        else:
            weights = self.data_object.get_selected_data(
                weight_key,
                sl=sl,
                *args,
                **data_kwargs['weight']
            ).flatten()

            if fix_invalid:
                weights = np.ma.masked_array( weights, mask=mask ).compressed()

        if x_range is default:
            x_range = [ x_data.min(), x_data.max() ]
        elif isinstance( x_range, float ):
            x_range = np.array( [ -x_range, x_range ])*self.data_object.length_scale[slices]
        if y_range is default:
            y_range = [ y_data.min(), y_data.max() ]
        elif isinstance( y_range, float ):
            y_range = np.array( [ -y_range, y_range ])*self.data_object.length_scale[slices]

        if x_scale == 'log':
            x_edges = np.logspace( np.log10( x_range[0] ), np.log10( x_range[1] ), n_bins )
        else:
            x_edges = np.linspace( x_range[0], x_range[1], n_bins )
        if y_scale == 'log':
            y_edges = np.logspace( np.log10( y_range[0] ), np.log10( y_range[1] ), n_bins )
        else:
            y_edges = np.linspace( y_range[0], y_range[1], n_bins )

        # Make the histogram
        hist2d, x_edges, y_edges = np.histogram2d( x_data, y_data, [x_edges, y_edges], weights=weights, normed=normed )

        # If doing an average, divide by the number in each bin
        if average:
            average_hist2d, x_edges, y_edges = np.histogram2d( x_data, y_data, [x_edges, y_edges], normed=normed )
            hist2d /= average_hist2d

        # If making the y-axis conditional, divide by the distribution of data for the x-axis.
        if conditional_y:
            hist_x, x_edges = np.histogram( x_data, x_edges, normed=normed )
            hist2d /= hist_x[:,np.newaxis]

        # Divide the histogram bins by this array
        if hist_div_arr is not None:
            hist2d /= hist_div_arr

        # Mask bins below a specified value
        if min_bin_value_displayed is not None:
            hist2d = np.ma.masked_where(
                hist2d < min_bin_value_displayed,
                hist2d,
            )

        # Plot
        if ax is default:
            fig = plt.figure( figsize=(10,9), facecolor='white' )
            ax = plt.gca()

        if z_scale == 'linear':
            norm = plt_colors.Normalize()
        elif z_scale == 'log':
            norm = plt_colors.LogNorm()

        if cdf:
            raise Exception(
                "Not implemented yet. When implementing, use utilities.cumsum2d"
            )

        im = ax.pcolormesh(
            x_edges,
            y_edges,
            hist2d.transpose(),
            cmap = cmap,
            norm = norm,
            vmin = vmin,
            vmax = vmax,
            zorder = zorder,
        )

        # Add a colorbar
        if add_colorbar:
            if colorbar_args is default:
                colorbar_args = ax
                cbar = gen_plot.add_colorbar( colorbar_args, im, method='ax' )
            else:
                colorbar_args['color_object'] = im
                cbar = gen_plot.add_colorbar( **colorbar_args )
            cbar.ax.tick_params( labelsize=20 )

        # Plot Line for easier visual interpretation
        if line_slope is not default:
            line_x = np.array( [ x_data.min(), x_data.max() ] )
            line_y = line_slope*line_x
            ax.plot( line_x, line_y, linewidth=3, linestyle='dashed', )

        if horizontal_line is not None:
            trans = transforms.blended_transform_factory( ax.transAxes, ax.transData )
            ax.plot( [ 0., 1. ], [ horizontal_line, ]*2, transform=trans, **horizontal_line_kwargs )
        if vertical_line is not None:
            trans = transforms.blended_transform_factory( ax.transData, ax.transAxes )
            ax.plot( [ vertical_line, ]*2, [ 0., 1. ], transform=trans, **vertical_line_kwargs )

        # Plot label
        if plot_label is not None:
            if plot_label is default:
                plt_label = ax.annotate(
                    s = self.label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, str ):
                plt_label = ax.annotate(
                    s = plot_label,
                    xy = (0.,1.0),
                    va = 'bottom',
                    xycoords = 'axes fraction',
                    fontsize = label_fontsize,
                 )
            elif isinstance( plot_label, dict ):
                plt_label = ax.annotate( **plot_label )
            else:
                raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
            if outline_plot_label:
                plt_label.set_path_effects([ path_effects.Stroke(linewidth=3, foreground='black'), path_effects.Normal() ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            try:
                info_label = r'$z=' + '{:.3f}'.format( self.data_object.redshift ) + '$'+ info_label
            except ValueError:
                info_label = r'$z=' + '{:.3f}'.format( self.data_object.redshift.values[sl[1]] ) + '$'+ info_label
        if label_galaxy_cut or label_redshift:
            ax.annotate( s=info_label, xy=(1.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,
                ha='right' )

        # Add axis labels
        if add_x_label:
            if x_label is default:
                x_label = x_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is default:
                y_label = y_key
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        # Limits
        ax.set_xlim( x_range )
        ax.set_ylim( y_range )

        # Scale
        ax.set_xscale( x_scale )
        ax.set_yscale( y_scale )

        # Set tick parameters
        if tick_param_args is not default:
            ax.tick_params( **tick_param_args )

        # Save the file
        if out_dir is not None:
            if save_file is default:
                save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )

            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            plt.close()

        # Return?
        if return_dist:
            return hist2d, x_edges, y_edges

    ########################################################################

    def scatter(
        self,
        x_key, y_key,
        slices = None,
        n_subsample = None,
        ax = default,
        marker_size = 100,
        color = 'k',
        marker = '.',
        zorder = -100,
        x_range = default, y_range = default,
        x_label = default, y_label = default,
        add_x_label = True, add_y_label = True,
        plot_label = default,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = False,
        label_fontsize = 24,
        tick_param_args = default,
        out_dir = None,
        fix_invalid = True,
        line_slope = default,
        *args, **kwargs ):
        '''Make a 2D scatter plot of the data. Extra arguments are passed to get_selected_data.
        Args:
            x_key, y_key (str) : Data keys to plot.
            weight_key (str) : Data key for data to use as a weight. By default, no weight.
            slices (int or tuple of slices) : How to slices the data.
            ax (axis) : What axis to use. By default creates a figure and places the axis on it.
            x_range, y_range ( (float, float) ) : Histogram edges. If default, all data is enclosed. If list, set manually.
                If float, is +- x_range*length scale at that snapshot.
            n_bins (int) : Number of bins in the histogram.
            vmin, vmax (float) : Limits for the colorbar.
            plot_halos (bool) : Whether or not to plot merger tree halos on top of the histogram.
                Only makes sense for when dealing with positions.
            add_colorbar (bool) : If True, add a colorbar to colorbar_args
            colorbar_args (axis) : What axis to add the colorbar to. By default, is ax.
            x_label, ylabel (str) : Axes labels.
            add_x_label, add_y_label (bool) : Include axes labels?
            plot_label (str or dict) : What to label the plot with. By default, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            tick_param_args (args) : Arguments to pass to ax.tick_params. By default, don't change inherent defaults.
            out_dir (str) : If given, where to save the file.
            fix_invalid (bool) : Fix invalid values.
            line_slope (float) : If given, draw a line with the given slope.
        '''

        if isinstance( slices, int ):
            sl = ( slice(None), slices )
        else:
            sl = slices

        # Get data
        x_data = self.data_object.get_selected_data( x_key, sl=sl, *args, **kwargs )
        y_data = self.data_object.get_selected_data( y_key, sl=sl, *args, **kwargs )

        # Fix NaNs
        if fix_invalid:
            x_mask = np.ma.fix_invalid( x_data ).mask
            y_mask = np.ma.fix_invalid( y_data ).mask
            mask = np.ma.mask_or( x_mask, y_mask )

            x_data = np.ma.masked_array( x_data, mask=mask ).compressed()
            y_data = np.ma.masked_array( y_data, mask=mask ).compressed()

        # Subsample
        if n_subsample is not None:
            sampled_inds = np.random.randint( 0, x_data.size, n_subsample )

            x_data = x_data[sampled_inds]
            y_data = y_data[sampled_inds]

        if x_range is default:
            x_range = [ x_data.min(), x_data.max() ]
        elif isinstance( x_range, float ):
            x_range = np.array( [ -x_range, x_range ])*self.data_object.ptracks.length_scale.iloc[slices]
        if y_range is default:
            y_range = [ y_data.min(), y_data.max() ]
        elif isinstance( y_range, float ):
            y_range = np.array( [ -y_range, y_range ])*self.data_object.ptracks.length_scale.iloc[slices]

        # Plot
        if ax is default:
            fig = plt.figure( figsize=(10,9), facecolor='white' )
            ax = plt.gca()

        s = ax.scatter( x_data, y_data, s=marker_size, color=color, marker=marker )

        # Change the z order
        s.set_zorder( zorder )

        # Halo Plot
        if line_slope is not default:
            line_x = np.array( [ x_data.min(), x_data.max() ] )
            line_y = line_slope*line_x
            ax.plot( line_x, line_y, linewidth=3, linestyle='dashed', )

        # Plot label
        if plot_label is default:
            plt_label = ax.annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, str ):
            plt_label = ax.annotate( s=plot_label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, dict ):
            plt_label = ax.annotate( **plot_label )
        elif plot_label is None:
            pass
        else:
            raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
        if outline_plot_label:
            plt_label.set_path_effects([ path_effects.Stroke(linewidth=3, foreground='black'), path_effects.Normal() ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            info_label = r'$z=' + '{:.3f}'.format( self.data_object.ptracks.redshift.iloc[slices] ) + '$, '+ info_label
        if label_galaxy_cut or label_redshift:
            ax.annotate( s=info_label, xy=(1.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,
                ha='right' )

        # Add axis labels
        if add_x_label:
            if x_label is default:
                x_label = x_key
            ax.set_xlabel( x_label, fontsize=label_fontsize )
        if add_y_label:
            if y_label is default:
                y_label = y_key
            ax.set_ylabel( y_label, fontsize=label_fontsize )

        # Limits
        ax.set_xlim( x_range )
        ax.set_ylim( y_range )

        # Set tick parameters
        if tick_param_args is not default:
            ax.tick_params( **tick_param_args )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )
            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            plt.close()

    ########################################################################

    def plot_stacked_data(
        self,
        x_key,
        y_keys,
        colors,
        ax = default,
        *args, **kwargs
    ):
        
        if ax is default:
            plt.figure( figsize=(11, 5), facecolor='white' )
            ax = plt.gca()

        y_prev = np.zeros( shape=y_datas.values()[0].shape )

        y_datas = []
        for y_key in y_keys:
            y_data = self.data_object.get_selected_data(
                y_key,
                *args, **kwargs
            ).copy()

            y_datas.append( y_data )

        for i, y_key in y_keys:

            y_next = y_prev + y_datas[i]

            ax.fill_between(
                x_data,
                y_prev,
                y_next,
                color = classification_colors[key],
                alpha = p_constants.CLASSIFICATION_ALPHA,
            )

            # Make virtual artists to allow a legend to appear
            color_object = matplotlib.patches.Rectangle(
                (0, 0),
                1,
                1,
                fc = classification_colors[key],
                ec = classification_colors[key],
                alpha = p_constants.CLASSIFICATION_ALPHA,
            )
            color_objects.append( color_object )
            labels.append( p_constants.CLASSIFICATION_LABELS[key] )

        ax.annotate(
            s=self.label,
            xy=(0., 1.0225),
            xycoords='axes fraction',
            fontsize=22,
        )

        ax.legend(
            color_objects,
            labels,
            prop={'size': 14.5},
            ncol=5,
            loc=(0., -0.28),
            fontsize=20
        )

    ########################################################################

    def plot_time_dependent_data( self,
        ax = default,
        x_range = [ 0., np.log10(8.) ], y_range = default,
        y_scale = 'log',
        x_label = default, y_label = default,
        ):
        '''Make a plot like the top panel of Fig. 3 in Angles-Alcazar+17

        Args:
            ax (axis object) :
                What axis to put the plot on. By default, create a new one on a separate figure.

            x_range, y_range (list-like) :
                [ x_min, x_max ] or [ y_min, y_max ] for the displayed range.

            x_label, y_label (str) :
                Labels for axis. By default, redshift and f(M_star), respectively.

            plot_dividing_line (bool) :
                Whether or not to plot a line at the edge between stacked regions.
        '''

        if ax is default:
            fig = plt.figure( figsize=(11,5), facecolor='white' )
            ax = plt.gca()

        x_data = np.log10( 1. + self.data_object.get_data( 'redshift' ) )

        y_data = self.data_object.get_categories_galaxy_mass()

        for key in p_constants.CLASSIFICATION_LIST_A[::-1]:

            y_data = y_datas[key]

            ax.plot(
                x_data,
                y_data,
                linewidth = 3,
                color = p_constants.CLASSIFICATION_COLORS[key],
                label = p_constants.CLASSIFICATION_LABELS[key],
            )

        if x_range is not default:
            ax.set_xlim( x_range )

        if y_range is not default:
            ax.set_ylim( y_range )

        ax.set_yscale( y_scale )

        tick_redshifts = np.array( [ 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, ] )
        x_tick_values = np.log10( 1. + tick_redshifts )
        plt.xticks( x_tick_values, tick_redshifts )

        ax.set_xlabel( r'z', fontsize=22, )
        ax.set_ylabel( r'$M_{\star} (M_{\odot})$', fontsize=22, )

        ax.annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=22,  )

        ax.legend( prop={'size':14.5}, ncol=5, loc=(0.,-0.28), fontsize=20 )

    ########################################################################
    # Generic Plotting Methods
    ########################################################################

    def same_axis_plot(
        self,
        axis_plotting_method_str,
        variations,
        ax = default,
        figsize = (11, 5),
        out_dir = None,
        add_line_label = False,
        legend_args = { 'prop': {'size': 16.5}, 'loc': 'upper right', 'fontsize': 20 },
        *args, **kwargs
    ):

        if ax is default:
            fig = plt.figure( figsize=figsize, facecolor='white', )
            ax = plt.gca()

        all_plotting_kwargs = utilities.dict_from_defaults_and_variations( kwargs, variations )

        axis_plotting_method = getattr( self, axis_plotting_method_str )
        for key, plotting_kwargs in all_plotting_kwargs.items():

            plotting_kwargs['ax'] = ax

            if add_line_label:
                plotting_kwargs['line_label'] = key

            axis_plotting_method( *args, **plotting_kwargs )

        ax.legend( **legend_args )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[kwargs['slices']] )
            gen_plot.save_fig( out_dir, save_file, fig=fig, dpi=75 )

            plt.close()

    ########################################################################

    def panel_plot( self,
        panel_plotting_method_str,
        defaults,
        variations,
        slices = default,
        n_rows = 2,
        n_columns = 2,
        plot_locations = [ (0,0), (0,1), (1,0), (1,1) ],
        figsize = (10,9),
        plot_label = default,
        outline_plot_label = False,
        label_galaxy_cut = False,
        label_redshift = True,
        label_fontsize = 24,
        subplot_label_args = { 'xy': (0.075, 0.88), 'xycoords': 'axes fraction', 'fontsize': 20, 'color': 'k',  },
        subplot_spacing_args = { 'hspace': 0.0001, 'wspace': 0.0001, },
        out_dir = None,
        ):
        '''
        Make a multi panel plot of the type of your choosing.

        Args:
            panel_plotting_method_str (str) : What type of plot to make.
            defaults (dict) : Default arguments to pass to panel_plotting_method.
            variations (dict of dicts) : Differences in plotting arguments per subplot.
            slices (slice) : What slices to select. By default, this doesn't pass any slices argument to panel_plotting_method
            plot_label (str or dict) : What to label the plot with. By default, uses self.label.
                Can also pass a dict of full args.
            outline_plot_label (bool) : If True, add an outline around the plot label.
            label_galaxy_cut (bool) : If true, add a label that indicates how the galaxy was defined.
            label_redshift (bool) : If True, add a label indicating the redshift.
            label_fontsize (int) : Fontsize for the labels.
            subplot_label_args (dict) : Label arguments to pass to each subplot for the label for the subplot.
                The actual label string itself corresponds to the keys in variations.
            subplot_spacing_args (dict) : How to space the subplots.
            out_dir (str) : If given, where to save the file.
        '''

        fig = plt.figure( figsize=figsize, facecolor='white', )
        ax = plt.gca()

        fig.subplots_adjust( **subplot_spacing_args )

        if slices is not default:
            defaults['slices'] = slices

        plotting_kwargs = utilities.dict_from_defaults_and_variations( defaults, variations )

        # Setup axes
        gs = gridspec.GridSpec(n_rows, n_columns)
        axs = []
        for plot_location in plot_locations:
            axs.append( plt.subplot( gs[plot_location] ) )

        # Setup arguments further
        for i, key in enumerate( plotting_kwargs.keys() ):
            ax_kwargs = plotting_kwargs[key]

            ax_kwargs['ax'] = axs[i]

            # Subplot label args
            this_subplot_label_args = subplot_label_args.copy()
            this_subplot_label_args['s'] = key
            ax_kwargs['plot_label'] = this_subplot_label_args

            if ax_kwargs['add_colorbar']:
                ax_kwargs['colorbar_args'] = { 'fig_or_ax': fig, 'ax_location': [0.9, 0.125, 0.03, 0.775 ],  }

            # Clean up interior axes
            ax_tick_parm_args = ax_kwargs['tick_param_args'].copy()
            plot_location = plot_locations[i]
            # Hide repetitive x labels
            if plot_location[0] != n_rows -1 :
                ax_kwargs['add_x_label'] = False
                ax_tick_parm_args['labelbottom'] = False
            # Hide repetitive y labels
            if plot_location[1] != 0:
                ax_kwargs['add_y_label'] = False
                ax_tick_parm_args['labelleft'] = False
            ax_kwargs['tick_param_args'] = ax_tick_parm_args

        # Actual panel plots
        panel_plotting_method = getattr( self, panel_plotting_method_str )
        for key in plotting_kwargs.keys():
            panel_plotting_method( **plotting_kwargs[key] )

        # Main axes labels
        # Plot label
        if plot_label is default:
            plt_label = axs[0].annotate( s=self.label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, str ):
            plt_label = axs[0].annotate( s=plot_label, xy=(0.,1.0225), xycoords='axes fraction', fontsize=label_fontsize,  )
        elif isinstance( plot_label, dict ):
            plt_label = axs[0].annotate( **plot_label )
        else:
            raise Exception( 'Unrecognized plot_label arguments, {}'.format( plot_label ) )
        if outline_plot_label:
            plt_label.set_path_effects([
                path_effects.Stroke(linewidth=3, foreground='white', background='white'),
                path_effects.Normal()
            ])

        # Upper right label (info label)
        info_label = ''
        if label_galaxy_cut:
            info_label = r'$r_{ \rm cut } = ' + '{:.3g}'.format( self.data_object.galids.parameters['galaxy_cut'] ) + 'r_{ s}$'
        if label_redshift:
            ind = defaults['slices']
            info_label = r'$z=' + '{:.3f}'.format( self.data_object.ptracks.redshift.iloc[ind] ) + '$'+ info_label
        if label_galaxy_cut or label_redshift:
            label_ax = plt.subplot( gs[0,n_columns-1,] )
            label_ax.annotate(
                s=info_label,
                xy=(1.,1.0225),
                xycoords='axes fraction',
                fontsize=label_fontsize,
                ha='right'
            )

        # Save the file
        if out_dir is not None:
            save_file = '{}_{:03d}.png'.format( self.label, self.data_object.ptracks.snum[slices] )
            gen_plot.save_fig( out_dir, save_file, fig=fig )

            plt.close()

    ########################################################################

    def make_multiple_plots( self,
        plotting_method_str,
        iter_args_key,
        iter_args,
        n_processors = 1,
        out_dir = None,
        make_out_dir_subdir = True,
        make_movie = False,
        clear_data = False,
        *args, **kwargs ):
        '''Make multiple plots of a selected type. *args and **kwargs are passed to plotting_method_str.

        Args:
            plotting_method_str (str) : What plotting method to use.
            iter_args_key (str) : The name of the argument to iterate over.
            iter_args (list) : List of argument values to change.
            n_processors (int) : Number of processors to use. Should only be used when saving the data.
            out_dir (str) : Where to save the data.
            make_movie (bool) : Make a movie out of the plots, if True.
            clear_data (bool) : If True, clear memory of the data after making the plots.
        '''

        plotting_method = getattr( self, plotting_method_str )

        if ( out_dir is not None ) and make_out_dir_subdir:
            out_dir = os.path.join( out_dir, self.label )

        def plotting_method_wrapper( process_args ):

            used_out_dir, used_args, used_kwargs = process_args

            plotting_method( out_dir=used_out_dir, *used_args, **used_kwargs )

            del used_out_dir, used_args, used_kwargs

            return

        all_process_args = []
        for iter_arg in iter_args:
            process_kwargs = dict( kwargs )
            process_kwargs[iter_args_key] = iter_arg
            all_process_args.append( ( out_dir, args, process_kwargs ) )

        if n_processors > 1:
            # For safety, make sure we've loaded the data already
            self.data_object.ptracks, self.data_object.galids, self.data_object.classifications

            mp_utils.parmap( plotting_method_wrapper, all_process_args, n_processors=n_processors, return_values=False )
        else:
            for i, iter_arg in enumerate( iter_args ):
                plotting_method_wrapper( all_process_args[i] )

        if make_movie:
            gen_plot.make_movie( out_dir, '{}_*.png'.format( self.label ), '{}.mp4'.format( self.label ), )

        if clear_data:
            del self.data_object.ptracks
            del self.data_object.galids
            del self.data_object.classifications

########################################################################
########################################################################

class PlotterSet( utilities.SmartDict ):
    '''Container for multiple plotters that is an enhanced dictionary.
    '''

    def __init__( self, data_object_cls, plotter_object_cls, defaults, variations ):
        '''
        Args:
            data_object_cls (object) : Class for the data object.
            plotter_object_cls (object) : Class for the plotter object.
            defaults (dict) : Set of default arguments for loading worldline data.
            variations (dict of dicts) : Labels and differences in arguments to be passed to Worldlines
        '''

        # Load the worldline sets
        storage = {}
        for key in variations.keys():

            kwargs = dict( defaults )
            for var_key in variations[key].keys():
                kwargs[var_key] = variations[key][var_key]

            storage[key] = { 'data_object': data_object_cls( **kwargs ), 'label': key }

        plotters_storage = utilities.SmartDict.from_class_and_args( plotter_object_cls, storage )

        super( PlotterSet, self ).__init__( plotters_storage )



import setuptools

setuptools.setup(
    name='galaxy_dive',
    version='0.8.2',
    description='A general analysis suite for hydrodynamic galaxy simulations.',
    url='https://github.com/zhafen/galaxy-dive',
    author='Zach Hafen',
    author_email='zachary.h.hafen@gmail.com',
    packages=setuptools.find_packages(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'numpy>=1.14.5',
        'pandas>=0.20.3',
        'mock>=2.0.0',
        'pytest>=3.4.0',
        'unyt>=1.0.4',
        'colossus>=1.2.1',
        'setuptools>=28.8.0',
        'matplotlib>=2.0.2',
        'h5py>=2.7.0',
        'scipy>=1.1.0',
    ],
    include_package_data=True,
)



#!/usr/bin/env python
'''Input and output

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import h5py
import numpy as np
import os

########################################################################
# Functions for dealing with files
########################################################################

# Create a copy of a hdf5 file and return it.

def copyHDF5(source_filename, copy_filename):

  # Open files
  f = h5py.File(source_filename, 'r')
  g = h5py.File(copy_filename, 'a')

  # Copy the contents
  for key in f:
    f.copy(key, g)
  for attr_key in f.attrs:
    g.attrs[attr_key] = f.attrs[attr_key]

  f.close()

  return g

########################################################################

# Add a character after the end of a general number of columns (this function is effectively made only to put MUSIC input files of points in a form legible by MUSIC, thus the default added character is a space).

def addCharEndLine(source_filename, char=' '):

  data = np.loadtxt(source_filename)

  np.savetxt(source_filename, data, fmt='%.6f', delimiter=char, newline=(char + '\n'))

  return 0

########################################################################
# Functions for dealing with labels and names 
########################################################################

# Get abbreviated names for simulations based off of their snapshot directory.

def abbreviatedName(snap_dir):

  # Dissect the name for the useful abbreviations.
  divided_sdir = snap_dir.split('/')
  if divided_sdir[-1] == '':
    del divided_sdir[-1]
  if divided_sdir[-1] == 'output':
    sim_name_index = -2
  else:
    sim_name_index = -1
  sim_name_full = divided_sdir[sim_name_index]
  sim_name_divided = sim_name_full.split('_')
  sim_name = sim_name_divided[0]

  # Attach the appropriate name, accounting for the sometimes weird naming conventions.
  if sim_name == 'B1':
    sim_name = 'm12v'
  elif sim_name == 'm12qq':
    sim_name = 'm12q'
  elif sim_name == 'm12v':
    sim_name = 'm12i'
  elif sim_name == 'massive':
    sim_name = 'MFz0_A2'
  elif sim_name == 'm12c':
    sim_name = 'm11.4a'
  elif sim_name == 'm12d':
    sim_name = 'm11.9a'

  # Check for alternate runs (e.g. turbdiff)
  if len(sim_name_divided) > 1:
    if sim_name_divided[-1] == 'turbdiff':
      sim_name += 'TD'
    elif sim_name_divided[1] == 'res7000':
      sim_name += '2'
    
      # Account for variants
      for addition in sim_name_divided[2:]:
        sim_name += addition

  return sim_name

########################################################################

# Get the abbreviated names for a list of simulations.

def abbreviated_name_list(snap_dir_list):
    # Loop over all simulations.
    sim_name_list = []
    for snap_dir in snap_dir_list:

      sim_name = abbreviatedName(snap_dir)

      sim_name_list.append(sim_name)

    return sim_name_list

########################################################################

# Get the ionization tag for grid and LOS filenames.

def ionTag(ionized):

  if ionized == False:
    ionization_tag = ''
    return ionization_tag

  elif ionized == 'RT':
    ionization_tag = '_ionized'
  elif ionized.split('_')[0] == 'R13':
    ionization_tag = '_' + ionized
  else:
    ionization_tag = '_' + ionized

  return ionization_tag

########################################################################

# Get the gridded snapshot filename from a few inputs

def getGridFilename(sim_dir, snap_id, Nx, gridsize, ionized, ion_grid=False):
# sim_dir: Simulation directory, string.
# snap_id: Snapshot ID, integer
# Nx: Grid resolution, integer
# gridsize: Grid size, float or string

  ionization_tag = ionTag(ionized)

  grid_filename = 'grid_%i_%i_%s_sasha%s.hdf5'%(snap_id, Nx, str(gridsize), ionization_tag)

  if ion_grid:
    grid_filename = 'ion_' + grid_filename
  
  grid_path = os.path.join(sim_dir, grid_filename)

  return grid_path

########################################################################

# Break apart gridded snapshot file names

def breakGridFilename(gridded_snapshot_file):

  # Break the simulation path apart
  path_to_file = gridded_snapshot_file.split('/')
  simulation = path_to_file[-2]
  grid_file = path_to_file[-1]

  # Break the grid file apart
  seperated_file = grid_file.split('_')
  snap_id = seperated_file[1]
  Nx = seperated_file[2]
  gridsize = seperated_file[3]

  return (simulation, snap_id, Nx, gridsize)

########################################################################

# Get the halofile name from a few inputs

def getHaloFilename(sim_dir, halo_number):
  
  return '{}/halo_{:0>5d}.datsimple.txtsmooth'.format(sim_dir, halo_number)

########################################################################

# Get the location of the LOS data from a few inputs.

def getLOSDataFilename(snap_dir, Nx, gridsize, face, comp_method, ionized=False, den_weight='nH'):

  # Simulation name
  snap_dir_divided = snap_dir.split('/')
  if snap_dir_divided[-1] == 'output':
    snap_dir_name = snap_dir_divided[-2]
  else:
    snap_dir_name = snap_dir.split('/')[-1]

  ionization_tag = ionTag(ionized)

  if den_weight == 'nH':
    den_weight_tag = ''
  else:
    den_weight_tag = '_' + den_weight

  return '/work/03057/zhafen/LOSdata/LOS_{}_{}_{}_{}_{}{}{}.hdf5'.format(snap_dir_name, Nx, str(gridsize), face, comp_method, ionization_tag, den_weight_tag)




# Adapted from Behroozi+2018 (UniverseMachine EDR) by S.Wellons 7/11/18

#!/usr/bin/python
import sys
import math
from scipy.optimize import newton
import numpy as np

def get_Mhalo(Mstar, z, paramfile="./params/smhm_true_med_params.txt"):
    Mhalo = np.zeros(len(Mstar))
    for i in range(0,len(Mstar)):
        Mhalo[i] = newton(_get_Mstar, 12., args=(z, Mstar[i]))
    return Mhalo

def _get_Mstar(Mhalo, z, target):
    return get_Mstar(np.asarray([Mhalo]), z, paramfile="smhm_true_med_params.txt")[0] - target
            

def get_Mstar(Mhalo, z, paramfile="./params/smhm_true_med_params.txt"):

    #Load params
    param_file = open(paramfile, "r") 
    param_list = []
    allparams = []
    for line in param_file:
        param_list.append(float((line.split(" "))[1]))
        allparams.append(line.split(" "))

    if (len(param_list) != 20):
        print("Parameter file not correct length.  (Expected 20 lines, got %d)." % len(param_list))
        quit()

    names = "EFF_0 EFF_0_A EFF_0_A2 EFF_0_Z M_1 M_1_A M_1_A2 M_1_Z ALPHA ALPHA_A ALPHA_A2 ALPHA_Z BETA BETA_A BETA_Z DELTA GAMMA GAMMA_A GAMMA_Z CHI2".split(" ");
    params = dict(zip(names, param_list))

    #Decide whether to print tex or evaluate SMHM parameter
    try:
        z = float(z)
    except:
        #print TeX
        for x in allparams[0:8:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write("\\\\\n & & & ")
        for x in allparams[8:16:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write("\\\\\n & & & ")    
        for x in allparams[16:19:1]:
            sys.stdout.write('& $%.3f^{+%.3f}_{-%.3f}$' % tuple(float(y) for y in x[1:4]))
        sys.stdout.write(' & %.0f' % float(allparams[19][1]))
        if (float(allparams[19][1])>200):
            sys.stdout.write('$\dag$')
        print('\\\\[2ex]')
        quit()

    #Print SMHM relation
    a = 1.0/(1.0+z)
    a1 = a - 1.0
    lna = math.log(a)
    zparams = {}
    zparams['m_1'] = params['M_1'] + a1*params['M_1_A'] - lna*params['M_1_A2'] + z*params['M_1_Z']
    zparams['sm_0'] = zparams['m_1'] + params['EFF_0'] + a1*params['EFF_0_A'] - lna*params['EFF_0_A2'] + z*params['EFF_0_Z']
    zparams['alpha'] = params['ALPHA'] + a1*params['ALPHA_A'] - lna*params['ALPHA_A2'] + z*params['ALPHA_Z']
    zparams['beta'] = params['BETA'] + a1*params['BETA_A'] + z*params['BETA_Z']
    zparams['delta'] = params['DELTA']
    zparams['gamma'] = 10**(params['GAMMA'] + a1*params['GAMMA_A'] + z*params['GAMMA_Z'])

    smhm_max = 14.5-0.35*z
    #print('#Log10(Mpeak/Msun) Log10(Median_SM/Msun) Log10(Median_SM/Mpeak)')
    #print('#Mpeak: peak historical halo mass, using Bryan & Norman virial overdensity.')
    #print('#Overall fit chi^2: %f' % params['CHI2'])
    # if (params['CHI2']>200):
    #     print('#Warning: chi^2 > 200 implies that not all features are well fit.  Comparison with the raw data (in data/smhm/median_raw/) is crucial.')
    Mstar = Mhalo*0.
    for i in range(len(Mhalo)): #m in [x*0.05 for x in range(int(10.5*20),int(smhm_max*20+1),1)]:
        m = Mhalo[i]
        dm = m-zparams['m_1'];
        dm2 = dm/zparams['delta'];
        sm = zparams['sm_0'] - math.log10(10**(-zparams['alpha']*dm) + 10**(-zparams['beta']*dm)) + zparams['gamma']*math.exp(-0.5*(dm2*dm2));
        Mstar[i] = sm
        #print("%.2f %.6f %.6f" % (m,sm,sm-m))

    return Mstar

# Returns dlogMstar/dlogMhalo
def get_slope(Mhalo, z, paramfile="./params/smhm_true_med_params.txt"):

    #Load params
    param_file = open(paramfile, "r") 
    param_list = []
    allparams = []
    for line in param_file:
        param_list.append(float((line.split(" "))[1]))
        allparams.append(line.split(" "))

    if (len(param_list) != 20):
        print("Parameter file not correct length.  (Expected 20 lines, got %d)." % len(param_list))
        quit()

    names = "EFF_0 EFF_0_A EFF_0_A2 EFF_0_Z M_1 M_1_A M_1_A2 M_1_Z ALPHA ALPHA_A ALPHA_A2 ALPHA_Z BETA BETA_A BETA_Z DELTA GAMMA GAMMA_A GAMMA_Z CHI2".split(" ");
    params = dict(zip(names, param_list))

    z = float(z)

    #Print SMHM relation
    a = 1.0/(1.0+z)
    a1 = a - 1.0
    lna = math.log(a)
    zparams = {}
    zparams['m_1'] = params['M_1'] + a1*params['M_1_A'] - lna*params['M_1_A2'] + z*params['M_1_Z']
    zparams['sm_0'] = zparams['m_1'] + params['EFF_0'] + a1*params['EFF_0_A'] - lna*params['EFF_0_A2'] + z*params['EFF_0_Z']
    zparams['alpha'] = params['ALPHA'] + a1*params['ALPHA_A'] - lna*params['ALPHA_A2'] + z*params['ALPHA_Z']
    zparams['beta'] = params['BETA'] + a1*params['BETA_A'] + z*params['BETA_Z']
    zparams['delta'] = params['DELTA']
    zparams['gamma'] = 10**(params['GAMMA'] + a1*params['GAMMA_A'] + z*params['GAMMA_Z'])

    smhm_max = 14.5-0.35*z
    slope = Mhalo*0.
    for i in range(len(Mhalo)): #m in [x*0.05 for x in range(int(10.5*20),int(smhm_max*20+1),1)]:
        m = Mhalo[i]
        dm = m-zparams['m_1'];
        
        term1 = (zparams['alpha']*10.**(zparams['beta']*dm)+zparams['beta']*10.**(zparams['alpha']*dm))/(10.**(zparams['beta']*dm) + 10.**(zparams['alpha']*dm))
        term2 = -zparams['gamma']*dm*math.exp(-(dm/zparams['delta'])**2/2.)/zparams['delta']**2
        slope[i] = term1 + term2

    return slope

    



'''Testing for hdf5_wrapper.py
'''

import h5py
from mock import patch
import numpy as np
import numpy.testing as npt
import os
import shutil
import unittest

import galaxy_dive.utils.hdf5_wrapper as hdf5_wrapper

########################################################################


class SaneEqualityArray(np.ndarray):
    '''Numpy array subclass that allows you to test if two arrays are equal.'''

    def __eq__(self, other):
            return (isinstance(other, np.ndarray) and self.shape == other.shape and np.allclose(self, other))


def sane_eq_array(list_in):
    '''Wrapper for SaneEqualityArray, that takes in a list.'''

    arr = np.array(list_in)

    return arr.view(SaneEqualityArray)

########################################################################


class TestHDF5Wrapper(unittest.TestCase):

    def setUp(self):

        self.filename = './tests/test.h5'
        self.copy_filename = './tests/test_copy.h5'

        self.test_object = hdf5_wrapper.HDF5Wrapper

        self.test_instance = self.test_object(self.filename)

        self.test_data = {
            'a': np.array([1., 2., 3., 4.]),
            'b': np.array([6., 4., 2., 8.]),
            'c': np.array([9., -1., -3., 4.]),
        }

        self.test_data_complex = {
            'a': np.array([1., 2., 3., 4.]),
            'b': np.array([ [6., 7., 8.], [4., 3., 2.], [-2., -3., -4.], [-8., -7., -6.], ]),
            'c': np.array([9., -1., -3., 4.]),
            'd': np.array(['dog', 'cat', 'rabbit', 'mouse']),
        }

    def tearDown(self):

        if os.path.isfile(self.filename):
            os.system( 'rm {}'.format(self.filename) )
        if os.path.isfile(self.copy_filename):
            os.system( 'rm {}'.format(self.copy_filename) )
        if os.path.isdir( './tests/data/output_data' ):
            os.system( 'rm -r ./tests/data/output_data' )

    ########################################################################

    def test_can_create_class(self):

        if self.test_instance is None:
            assert False

    ########################################################################

    def test_fragile_makes_hdf5_file_blank(self):
        '''Part of the testing suite that will fail if anything else does.'''

        self.test_instance.clear_hdf5_file()

        # Make a file
        f = h5py.File(self.filename, 'a')
        try:
            f.create_dataset('a', self.test_data['a'])
        except RuntimeError:
            pass
        f.close()

        self.test_instance.clear_hdf5_file()

        f = h5py.File(self.filename, 'r')
        self.assertEqual(len(f.keys()), 0)
        f.close()

    ########################################################################

    def test_dataset_from_empty(self):
        '''Test that we can make a dataset from nothing.'''

        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(self.test_data)

        f = h5py.File(self.filename, 'r')
        npt.assert_allclose(self.test_data['a'], f['a'])
        f.close()

    ########################################################################

    def test_sort_dictionary(self):

        actual = self.test_instance.sort_dictionary(self.test_data, 'c')

        expected = {'a': np.array([3., 2., 4., 1.]),
                                'b': np.array([2., 4., 8., 6.]),
                                'c': np.array([-3., -1., 4., 9.]),
                              }

        for key in expected:
            npt.assert_allclose(actual[key], expected[key])

    ########################################################################

    def test_insert_data(self):
        '''Test that we can add to an existing dataset.'''

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(self.test_data)

        data_to_insert = {'a': 2.5, 'b': 4., 'c': 6.}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([1., 2., 2.5, 3., 4.]),
                                            'b': np.array([6., 4., 4., 2., 8.]),
                                            'c': np.array([9., -1., 6., -3., 4.]),
                                          }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_insert_data_complex(self):
        '''Test that we can add to an existing dataset, for a slightly more complicated dataset.'''

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(self.test_data_complex)

        data_to_insert = {'a': 2.5, 'b': np.array([1., -1., 5.]), 'c': 6., 'd': 'grimalkin'}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([1., 2., 2.5,  3., 4.]),
                                            'b': np.array([ [6., 7., 8.], [4., 3., 2.], [1., -1., 5.], [-2., -3., -4.], [-8., -7., -6.], ]),
                                            'c': np.array([9., -1., 6., -3., 4.]),
                                            'd': np.array(['dog', 'cat', 'grimalkin', 'rabbit', 'mouse']),
                                          }

        for key in expected_data:
            if key != 'd':
                npt.assert_allclose(expected_data[key], f[key])

        # Do the testing for when it's strings instead.
        for i, animal in enumerate(expected_data['d']):
            self.assertEqual(animal, f['d'][i])

        f.close()

    ########################################################################

    def test_insert_duplicate_data(self):
        '''Test that adding existing values to a dataset doesn't do anything..'''

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(self.test_data)

        data_to_insert = {'a': 3., 'b': 2., 'c': -3.}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([1., 2., 3., 4.]),
                                            'b': np.array([6., 4., 2., 8.]),
                                            'c': np.array([9., -1., -3., 4.]),
                                          }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_insert_data_empty_dataset(self):
        '''Test that we can use insert on an empty dataset.'''

        self.test_instance.clear_hdf5_file()

        data_to_insert = {'a': 2.5, 'b': 4., 'c': 6.}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([ 2.5, ]),
                                            'b': np.array([ 4., ]),
                                            'c': np.array([ 6., ]),
                                          }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_insert_data_no_file(self):
        '''Test that we can use insert on a file that doesn't exist yet.'''

        # Delete the file
        os.system( 'rm {}'.format(self.filename) )

        data_to_insert = {'a': 2.5, 'b': 4., 'c': 6.}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([ 2.5, ]),
                                            'b': np.array([ 4., ]),
                                            'c': np.array([ 6., ]),
                                          }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_insert_data_only_one_row_exists(self):

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        test_data = {'a': 2.,
                                            'b': 4.,
                                            'c': -1.,
                                          }

        self.test_instance.save_data(test_data)

        data_to_insert = {'a': 1., 'b': 6., 'c': 9.}
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key)

        f = h5py.File(self.filename, 'r')
        expected_data = {'a': np.array([1., 2., ]),
                                            'b': np.array([6., 4., ]),
                                            'c': np.array([9., -1., ]),
                                          }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_insert_data_column(self):
        '''Test that we can insert columns of data with insert data.
        '''

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        test_data = {
            'a': [ 2, 3, ],
            'b': [ 4., 5., ],
            'c': [ -1., 0., ],
        }

        self.test_instance.save_data(test_data)

        data_to_insert = {'a': [3, 2], 'd': [ 6., 7., ] }
        index_key = 'a'

        self.test_instance.insert_data(data_to_insert, index_key, True)

        f = h5py.File(self.filename, 'r')
        expected_data = {
            'a': np.array([2, 3, ]),
            'b': np.array([4., 5., ]),
            'c': np.array([-1., 0., ]),
            'd': np.array([7., 6., ]),
        }

        for key in expected_data:
            npt.assert_allclose(expected_data[key], f[key])
        f.close()

    ########################################################################

    def test_copy_data(self):

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(self.test_data)

        self.test_instance.copy_hdf5_file(self.copy_filename)

        assert os.path.isfile(self.copy_filename)

        os.system( 'rm {}'.format(self.copy_filename) )

    ########################################################################

    @patch('numpy.random.randint')
    def test_subsample(self, mock_randint):

        # Make a side effect for the random integer
        def side_effect(low, high, number):
            return np.array([3, 5, 7, 4])
        mock_randint.side_effect = side_effect

        subsample_data = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(subsample_data)

        result = self.test_instance.subsample_hdf5_file(2,)

        mock_randint.assert_called_once_with(0, 9, 4)

        expected = {
            'a': np.array( [ [4., 6.], [7., 5.] ] ),
            'b': np.array( [ [2., 3.], [10., 1.] ] ),
            }

        for key in result:
            npt.assert_allclose(expected[key], result[key])

    ########################################################################

    @patch('numpy.random.randint')
    def test_copy_data_subsample(self, mock_randint):

        # Make a side effect for the random integer
        def side_effect(low, high, number):
            return np.array([3, 5, 7, 4])
        mock_randint.side_effect = side_effect

        subsample_data = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(subsample_data)

        # Make the copy
        subsamples = 2
        self.test_instance.copy_hdf5_file(self.copy_filename, subsamples=subsamples)

        expected = {
            'a': np.array( [ [4., 6.], [7., 5.] ] ),
            'b': np.array( [ [2., 3.], [10., 1.] ] ),
            }


        g = h5py.File(self.copy_filename, 'r')
        for key in g.keys():
            npt.assert_allclose(expected[key], g[key])

        os.system( 'rm {}'.format(self.copy_filename) )

    ########################################################################

    def test_save_complicated_structure(self):

        group_1 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }
        group_2 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }
        group_3 = {},
        data = {
            'Group1': group_1,
            'Group2': group_2,
            'Group3': group_3,
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(data)

        f = h5py.File(self.filename, 'r')
        for group in ['Group1', 'Group2']:
            for key in f[group].keys():
                npt.assert_allclose( f[group][key][...], data[group][key] )

    ########################################################################

    @patch('numpy.random.randint')
    def test_copy_data_subsample_complicated_structure(self, mock_randint):

        # Make a side effect for the random integer
        def side_effect(low, high, number):
            return np.array([3, 5, 7, 4])
        mock_randint.side_effect = side_effect

        group_1 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }
        group_2 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
        }
        group_3 = {},
        subsample_data = {
            'Group1': group_1,
            'Group2': group_2,
            'Group3': group_3,
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(subsample_data)

        # Make the copy
        subsamples = 2
        self.test_instance.copy_hdf5_file(self.copy_filename, subsamples=subsamples)

        expected = {
            'a': np.array( [ [4., 6.], [7., 5.] ] ),
            'b': np.array( [ [2., 3.], [10., 1.] ] ),
            }


        g = h5py.File(self.copy_filename, 'r')
        for group in ['Group1', 'Group2']:
            for key in g[group].keys():
                npt.assert_allclose( g[group][key][...], expected[key] )

        os.system( 'rm {}'.format(self.copy_filename) )

    ########################################################################

    @patch('numpy.random.randint')
    def test_copy_data_subsample_particle_like_data(self, mock_randint):

        # Make a side effect for the random integer
        def side_effect(low, high, number):
            return np.array([3, 5, 2, 4])
        mock_randint.side_effect = side_effect

        group_1 = {
            'a': np.array( [ 1., 2., 3., 4., 5., 6.,  ] ),
            'b': np.array( [ [3., -9., 7., 2., 1., 3.], [7., 10., 2., 5., 3., 1.] ] ).transpose(),
            'attrs': { 'attrs_exist': True, },
        }
        group_2 = {
            'a': np.array( [ 1., 2., 3., 4., 5., 6.,  ] ),
            'b': np.array( [ [3., -9., 7., 2., 1., 3.], [7., 10., 2., 5., 3., 1.] ] ).transpose(),
            'attrs': { 'attrs_exist': True, },
        }
        group_3 = {
            'attrs': { 'attrs_exist': True, },
        }
        subsample_data = {
            'Group1': group_1,
            'Group2': group_2,
            'Group3': group_3,
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(subsample_data)

        # Make the copy
        subsamples = 4
        self.test_instance.copy_hdf5_file(self.copy_filename, subsamples=subsamples, particle_data=True)

        expected = {
            'a': np.array( [ 4., 6., 3., 5. ] ),
            'b': np.array( [ [2., 3., 7., 1.], [5., 1., 2., 3.] ] ).transpose(),
            }


        g = h5py.File(self.copy_filename, 'r')
        for group in ['Group1', 'Group2']:
            for key in g[group].keys():
                npt.assert_allclose( g[group][key][...], expected[key] )

        os.system( 'rm {}'.format(self.copy_filename) )

    ########################################################################

    def test_copy_and_subsample_particle_snapshot( self ):

        filedir = './tests/data/sdir/output'
        snum = 600
        subsamples = 2

        copy_filedir = './tests/data/output_data/'

        # Set the random seed
        np.random.seed(1234)

        hdf5_wrapper.copy_snapshot( filedir, snum, copy_filedir, subsamples )

        f = h5py.File( os.path.join(copy_filedir, 'snapdir_600/snapshot_600.0.hdf5'), 'r' )

        actual = f['PartType0']['ParticleIDs'][...]

        expected = np.array([50743870, 66266902])
        npt.assert_allclose( actual, expected )

    ########################################################################

    def test_save_data_group_attrs(self):

        group_1 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
            'attrs': {'attrs_exist': True},
        }
        group_2 = {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
            'attrs': {'attrs_exist': True},
        }
        group_3 = {
            'attrs': {'attrs_exist': True},
        }
        data = {
            'Group1': group_1,
            'Group2': group_2,
            'Group3': group_3,
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(data)

        f = h5py.File(self.filename, 'r')
        for group in ['Group1', 'Group2', 'Group3']:
            for key in f[group].keys():
                if key != 'attrs':
                    npt.assert_allclose( f[group][key][...], data[group][key] )
                else:
                    for attr_key in f[group].attrs.keys():
                        assert data[group][key][attr_key] == f[group].attrs[attr_key]

        os.system( 'rm {}'.format(self.copy_filename) )

    ########################################################################

    @patch('numpy.random.randint')
    def test_copy_data_subsample_attrs_included(self, mock_randint):

        # Make a side effect for the random integer
        def side_effect(low, high, number):
            return np.array([3, 5, 7, 4])
        mock_randint.side_effect = side_effect

        subsample_data = {
        'Group1': {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
            'attrs': { 'attrs_exist': True, },
            },
        'Group2': {
            'a': np.array( [ [1., 2., 3.], [4., 5., 6.], [6., 7., 8.] ] ),
            'b': np.array( [ [3., -9., 7.], [2., 1., 3.], [7., 10., 2.] ] ),
            'attrs': { 'attrs_exist': True, }
            },
        }

        # Make the original dataset
        self.test_instance.clear_hdf5_file()
        self.test_instance.save_data(subsample_data)

        # Make the copy
        subsamples = 2
        self.test_instance.copy_hdf5_file(self.copy_filename, subsamples=subsamples)

        expected = {
            'a': np.array( [ [4., 6.], [7., 5.] ] ),
            'b': np.array( [ [2., 3.], [10., 1.] ] ),
            }

        g = h5py.File(self.copy_filename, 'r')
        for group in g.keys():
            for key in g[group].keys():
                npt.assert_allclose(expected[key], g[group][key])

            assert g[group].attrs['attrs_exist'] == True

        os.system( 'rm {}'.format(self.copy_filename) )

########################################################################

class TestCopySnapshot( unittest.TestCase ):

    def setUp(self):

        self.kwargs = {
            'sdir': './tests/data/sdir/output',
            'snum': 600,
            'n_files': 8,
            'copy_dir': './tests/data/temp_sdir/output',
            'redistribute': True,
            }

    ########################################################################

    def tearDown( self ):

        if os.path.isdir( './tests/data/temp_sdir' ):
            shutil.rmtree( './tests/data/temp_sdir' )

    ########################################################################

    def test_copy_snapshot_and_redistribute( self ):

        # Set the random seed
        np.random.seed(1234)

        hdf5_wrapper.copy_snapshot( **self.kwargs )

        f = h5py.File( os.path.join( self.kwargs['copy_dir'], 'snapdir_600/snapshot_600.0.hdf5'), 'r' )

        expected_ids = {
            'PartType0'  : np.array([56577266,  2447814,  5836451, 48854626, 38697472, 26844748, 12518198, 13929795]),
            'PartType1': np.array([ 96490067, 111969149,  82963761, 105912179, 130202505,  99168777, 81451547,  86626678]),
            'PartType2': np.array([143876937, 142690251, 141208178, 142867765, 146105043, 144348523, 143790176, 144671722]),
            'PartType4': np.array([ 8240338, 33395336, 50845628, 53558278,  8091731, 33163814, 33193203, 37236498]),
        }

        expected_positions = {
            'PartType0': np.array([[ 43147.37910547,  40849.73566323,  42018.36946875,  41953.76112891,
                42250.75959483,  42921.24354077,  41515.60304779,  41709.71414106],
            [ 43788.15997097,  43738.97513012,  44166.32677499,  46120.83087771,
                44294.84369966,  46882.80661966,  44056.62971023,  40841.09006986],
            [ 45874.27118392,  45494.31487345,  46320.23228283,  46816.0623442 ,
                46870.14027289,  47536.57624099,  46369.82810828,  45640.13294909]]),
            'PartType1': np.array([[ 41875.66483633,  41871.32568498,  40495.20591946,  41840.55990359,
                42598.06746686,  42654.54906594,  41768.28244483,  41861.53839883],
            [ 44118.35014121,  44072.60902362,  43457.33636481,  45991.11822595,
                44444.10578658,  46092.61460387,  44187.48020816,  44126.8824762 ],
            [ 46261.71664927,  46307.86194507,  45084.47439923,  46928.75842962,
                46494.5582603 ,  46606.12689646,  46387.78823959,  46145.42036851]]),
            'PartType2': np.array([[ 42718.6585258 ,  42702.82772914,  41160.28685959,  42389.72019772,
                69109.89212334,  40110.83523464,  41484.2452451 ,  14234.55600016],
            [ 46788.61036456,  46854.27699445,  43801.19749786,  41226.63322432,
                31753.93759704,  40179.45607042,  42103.70372868,  15006.74252232],
            [ 47164.62123673,  47055.86509581,  47683.73835005,  46284.58074773,
                19872.44520005,  43478.28376903,  44507.59821399,  70046.93697334]]),
            'PartType4': np.array([[ 41882.74558912,  41884.14086056,  41867.59334168,  41875.93178495,
                41875.21304509,  41876.46763611,  41874.55928862,  41875.64058087],
            [ 44120.2944586 ,  44123.60968022,  44116.0175617 ,  44122.7025734 ,
                44123.03437806,  44121.16050547,  44124.66589757,  44121.83243625],
            [ 46253.94805193,  46252.28734946,  46256.24757957,  46257.53836431,
                46257.84478371,  46257.60302959,  46258.12604825,  46257.58911288]]),
        }

        # Correct for hubble...
        for key in expected_positions.keys():
            expected_positions[key] *= 0.70199999999999996

        for i in range( self.kwargs['n_files'] ):
            orig_filename = 'snapdir_600/snapshot_600.0.hdf5'
            orig_filepath = os.path.join( self.kwargs['sdir'], orig_filename )
            filename = 'snapdir_600/snapshot_600.{}.hdf5'.format( i )
            out_filepath = os.path.join( self.kwargs['copy_dir'], filename )

            actual = h5py.File( out_filepath, 'r' )
            expected = h5py.File( orig_filepath, 'r' )

            # Compare headers
            for key in expected['Header'].attrs.keys():
                if key == 'NumPart_ThisFile':
                    expected_ = np.array( [ 5, 5, 5, 0, 5, 0, ] )
                elif key == 'NumFilesPerSnapshot':
                    expected_ = self.kwargs['n_files']
                else:
                    expected_ = expected['Header'].attrs[key]
                actual_ = actual['Header'].attrs[key]
                if isinstance( expected_, np.ndarray ):
                    npt.assert_allclose( expected_, actual_ )
                else:
                    self.assertEqual( expected_, actual_ )

            # Compare values
            for ptype in expected_ids.keys():

                expected_id = expected_ids[ptype][i]
                actual_id = actual[ptype]['ParticleIDs'][...][0]
                npt.assert_allclose( expected_id, actual_id )

                expected_pos = expected_positions[ptype][:,i]
                actual_pos = actual[ptype]['Coordinates'][...][0]
                npt.assert_allclose( expected_pos, actual_pos )





#!/usr/bin/env python
'''Means to associate particles with galaxies and halos at any given time.

@author: Zach Hafen
@contact: zachary.h.hafen@gmail.com
@status: Development
'''

import copy
import numpy as np
import scipy.spatial

import unyt

import galaxy_dive.analyze_data.halo_data as h_data
import galaxy_dive.utils.utilities as utilities

########################################################################
########################################################################


class GalaxyFinder( object ):
    '''Find the association with galaxies and halos for a given set of particles
    at a given redshift.
    '''

    @utilities.store_parameters
    def __init__(
        self,
        particle_positions,
        redshift,
        snum,
        hubble,
        galaxy_cut,
        length_scale,
        mt_length_scale = None,
        halo_length_scale = 'Rvir',
        particle_masses = None,
        minimum_criteria = 'n_star',
        minimum_value = 10,
        ids_to_return = None,
        halo_data = None,
        halo_finder = 'AHF',
        halo_data_dir = None,
        halo_file_tag = None,
        mtree_halos_index = None,
        main_mt_halo_id = None,
        low_memory_mode = True,
        memory_mode_divisions = 10,
    ):
        '''Initialize.

        Args:
            particle_positions (array-like) :
                Positions with dimensions (n_particles, 3).
                If given a np.ndarray, assumes pkpc units.
                Also accepts a unyt array.

            redshift (float) :
                Redshift the particles are at.

            snum (int) :
                Snapshot the particles correspond to.

            hubble (float) :
                Cosmological hubble parameter (little h)

            galaxy_cut (float) :
                The fraction of the length scale a particle must be inside to be
                counted as part of a galaxy.

            length_scale (str) :
                Anything within galaxy_cut*length_scale is counted as being
                inside the galaxy.

            mt_length_scale (str) :
                Same as length scale, but for merger tree halos. Defaults to
                length_scale.

            halo_length_scale (str) :
                Anything within halo_length_scale is counted as being
                inside the halo.

            particle_masses (np.ndarray, optional) :
                Masses of particles.

            minimum_criteria (str) :
                Options...
                'n_star' -- halos must contain a minimum number of stars to
                    count as containing a galaxy.
                'M_star' -- halos must contain a minimum stellar mass to count
                    as containing a galaxy.

            minimum_value (int or float) :
                The minimum amount of something (specified in minimum criteria)
                in order for a galaxy to count as hosting a halo.

            ids_to_return (list of strs, optional) :
                The types of id you want to get out.

            halo_data (HaloData object, optional) :
                An instance of an object that retrieves halo data
                If not given initiate one using the halo_data_dir in kwargs

            halo_data_dir (str, optional) :
                Directory the AHF data is in. Necessary if halo_data is not
                provided.

            halo_finder (str, optional) :
                What code were the halos found with? Examples are 'AHF',
                'Rockstar'.

            halo_file_tag (int, optional) :
                What identifying tag to use for the merger tree files? Necessary
                if using merger tree information.

            mtree_halos_index (str or int, optional)  :
                The index argument to pass to AHFReader.get_mtree_halos().
                For most cases this should be the final snapshot number, but see
                AHFReader.get_mtree_halos's documentation.

            main_mt_halo_id (int, optional) :
                Index of the main merger tree halo.

            low_memory_mode (bool) :
                If True, use less memory at the cost of reduced performance.

            memory_mode_divisions (int) :
                When low_memory_mode is True, we reduce the memory cost by doing
                less work at once. This divides certain
                arrays into a number of divisions equal to
                memory_mode_divisions, and then the results are calculated for
                each division before bringing everything together.
        '''

        if self.mt_length_scale is None:
            self.mt_length_scale = length_scale

        # Setup the default halo_data
        if halo_data is None:
            self.halo_data = h_data.HaloData(
                self.halo_data_dir,
                halo_finder = self.halo_finder,
                mt_kwargs = { 
                    'index' : self.mtree_halos_index,
                    'tag' : self.halo_file_tag,
                },
        )

        if not isinstance( particle_positions, unyt.array.unyt_array ):
            particle_positions = copy.copy( particle_positions ) * unyt.kpc

        # In the case of a minimum stellar mass, we need to divide the minimum
        # value by 1/h when getting its values out.
        if self.minimum_criteria in [ 'M_star', 'Mvir']:
            self.min_conversion_factor = self.hubble
        else:
            self.min_conversion_factor = 1

        # Derived properties
        self.n_particles = self.particle_positions.shape[0]

    ########################################################################
    # General Properties Unique to an Instance
    ########################################################################

    @property
    def valid_halo_inds( self ):
        '''
        Returns:
            valid_halo_inds (np.ndarray) :
                Indices of *AHF_halos halos that satisfy our minimum criteria
                for containing a galaxy.
        '''

        if not hasattr( self, '_valid_halo_inds' ):

            # Apply a cut on containing a minimum amount of stars
            min_criteria = self.halo_data.get_data(
                self.minimum_criteria,
                snum = self.snum,
            )
            has_minimum_value = \
                min_criteria / self.min_conversion_factor >= self.minimum_value

            # Figure out which indices satisfy the criteria and choose only
            # those halos
            self._valid_halo_inds = np.where( has_minimum_value )[0]

        return self._valid_halo_inds

    ########################################################################

    @property
    def dist_to_all_valid_halos( self ):
        '''
        Returns:
            dist_to_all_valid_halos (np.ndarray) :
                Distance between self.particle_positions and all *.AHF_halos
                halos containing a galaxy (in pkpc).
        '''

        if not hasattr( self, '_dist_to_all_valid_halos' ):

            self._dist_to_all_valid_halos = \
                self.dist_to_all_valid_halos_fn( self.particle_positions )

        return self._dist_to_all_valid_halos

    def dist_to_all_valid_halos_fn( self, particle_positions ):
        '''
        Args:
            particle_positions (np.ndarray) :
                Location of particles to find the distance for.

        Returns:
            dist_to_all_valid_halos (np.ndarray) :
                Distance between the particle positions and all *.AHF_halos
                halos containing a galaxy (in pkpc).
        '''

        # Get the halo positions
        halo_pos_comov = np.array([
            self.halo_data.get_data( 'X', self.snum, units='kpc' ),
            self.halo_data.get_data( 'Y', self.snum, units='kpc' ),
            self.halo_data.get_data( 'Z', self.snum, units='kpc' ),
        ]).transpose()
        halo_pos = halo_pos_comov / ( 1. + self.redshift ) / self.hubble
        halo_pos_selected = halo_pos[self.valid_halo_inds]

        # Get the distances
        # Output is ordered such that dist[:,0] is the distance to the center of
        # halo 0 for each particle
        return scipy.spatial.distance.cdist(
            particle_positions, halo_pos_selected )

    ########################################################################

    @property
    def ahf_halos_length_scale_pkpc( self ):
        '''Actual values of the length scale used, e.g. Rvir.'''

        if not hasattr( self, '_ahf_halos_length_scale_pkpc' ):

            # Get the relevant length scale
            if self.length_scale == 'r_scale':
                # Get the scale radius
                r_vir = self.halo_data.get_data(
                    self.halo_length_scale,
                    self.snum,
                    units = 'kpc',
                )
                length_scale = r_vir / self.halo_data.get_data(
                    'cAnalytic',
                    self.snum,
                    )
            else:
                length_scale = self.halo_data.get_data(
                    self.length_scale,
                    self.snum,
                    units = 'kpc',
                )
            self._ahf_halos_length_scale_pkpc = \
                length_scale / ( 1. + self.redshift ) / self.hubble

        return self._ahf_halos_length_scale_pkpc

    ########################################################################
    # ID Finding Routines
    ########################################################################

    def find_ids( self ):
        '''Find relevant halo and galaxy IDs.

        Returns:
            galaxy_and_halo_ids (dict): Keys are...
            Parameters:
                halo_id (np.array of ints):
                    ID of the least-massive halo the particle is part of.
                host_halo_id (np.array of ints):
                    ID of the host halo the particle is part of.
                gal_id (np.array of ints):
                    ID of the smallest galaxy the particle is part of.
                host_gal_id (np.array of ints):
                    ID of the host galaxy the particle is part of.
                mt_halo_id (np.array of ints):
                    Merger tree ID of the least-massive halo the particle is
                    part of.
                mt_gal_id (np.array of ints):
                    Merger tree ID of the smallest galaxy the particle is
                    part of.
        '''

        # Dictionary to store the data in.
        galaxy_and_halo_ids = {}

        # Typically halo files aren't created for the first snapshot.
        # Account for this.
        if self.snum == 0:
            for id_type in self.ids_to_return:
                galaxy_and_halo_ids[id_type] = np.empty( self.n_particles )
                galaxy_and_halo_ids[id_type].fill( -2. )

            return galaxy_and_halo_ids

        # Halo ID works well, but not all the time.
        # In particular, it doesn't work when the length scale used is not
        # the virial radius. Break when this happens
        returning_a_halo_id = False
        for data_type in self.ids_to_return:
            if 'halo' in data_type:
                returning_a_halo_id = True
        diff_length_scale = (
            ( self.length_scale != 'Rvir' ) or
            ( self.mt_length_scale != 'Rvir' )
        )
        halo_id_broken = returning_a_halo_id and diff_length_scale
        assert not halo_id_broken, "Cannot currently return a halo ID when" + \
            " using a different length scale"

        # Actually get the data
        for id_type in self.ids_to_return:

            if id_type == 'halo_id':
                galaxy_and_halo_ids['halo_id'] = self.find_halo_id()
            elif id_type == 'host_halo_id':
                galaxy_and_halo_ids['host_halo_id'] = self.find_host_id()
            elif id_type == 'gal_id':
                galaxy_and_halo_ids['gal_id'] = \
                    self.find_halo_id( self.galaxy_cut )
            elif id_type == 'host_gal_id':
                galaxy_and_halo_ids['host_gal_id'] = \
                    self.find_host_id( self.galaxy_cut )
            elif id_type == 'mt_halo_id':
                galaxy_and_halo_ids['mt_halo_id'] = \
                    self.find_halo_id( type_of_halo_id='mt_halo_id' )
            elif id_type == 'mt_gal_id':
                galaxy_and_halo_ids['mt_gal_id'] = self.find_halo_id(
                    self.galaxy_cut, type_of_halo_id='mt_halo_id' )
            elif id_type == 'd_gal':
                galaxy_and_halo_ids['d_gal'] = \
                    self.find_d_gal()
            elif id_type == 'd_other_gal':
                galaxy_and_halo_ids['d_other_gal'] = \
                    self.find_d_other_gal()
            elif id_type == 'd_other_gal_scaled':
                galaxy_and_halo_ids['d_other_gal_scaled'] = \
                    self.find_d_other_gal( scaled=True )
            else:
                raise Exception( "Unrecognized id_type, {}".format( id_type ) )

        return galaxy_and_halo_ids

    ########################################################################

    def find_d_gal( self ):
        '''Find the distance to the center of the closest halo that contains a
        galaxy.

        Returns:
            d_gal (np.ndarray) : For particle i, d_gal[i] is the distance in
            pkpc to the center of the nearest galaxy.
        '''

        # Handle when no halos exist.
        if self.halo_data.get_n_halos( self.snum ) == 0:
             return -2. * np.ones( (self.n_particles,) )

        return np.min( self.dist_to_all_valid_halos, axis=1 )

    ########################################################################

    def find_d_other_gal( self, scaled=False ):
        '''Find the distance to the center of the closest halo that contains a
        galaxy, other than the main galaxy.

        Args:
            scaled (bool) : If True, scale d_other_gal by length_scale

        Returns:
            d_other_gal (np.ndarray) :
                For particle i, d_other_gal[i] is the distance in pkpc to the
                center of the nearest galaxy, besides the main galaxy.
        '''

        # Handle when no halos exist.
        if self.halo_data.get_n_halos( self.snum ) == 0:
             return -2. * np.ones( (self.n_particles,) )

        # Handle when all the halos aren't massive enough
        if self.valid_halo_inds.size == 0:
            return -2. * np.ones( (self.n_particles,) )

        self.halo_data.data_reader.get_mtree_halos(
            self.mtree_halos_index, self.halo_file_tag )

        mtree_halo = self.halo_data.data_reader.mtree_halos[ self.main_mt_halo_id ]

        if self.snum < mtree_halo.index.min():
            # This mimics what would happen if ind_main_gal wasn't
            # in self.valid_halo_inds
            ind_main_gal_in_valid_inds = np.array( [] )
        else:
            # The indice for the main galaxy is the same as the AHF_halos
            # ID for it.
            ind_main_gal = mtree_halo['ID'][ self.snum ]

            valid_halo_ind_is_main_gal_ind = \
                self.valid_halo_inds == ind_main_gal
            ind_main_gal_in_valid_inds = \
                np.where( valid_halo_ind_is_main_gal_ind )[0]

        # If the there's no index because it's automatically okay
        if ind_main_gal_in_valid_inds.size == 0:
            dist_to_all_valid_other_gals = self.dist_to_all_valid_halos
            valid_halo_inds_sats = self.valid_halo_inds

        elif ind_main_gal_in_valid_inds.size == 1:

            dist_to_all_valid_other_gals = np.delete(
                self.dist_to_all_valid_halos,
                ind_main_gal_in_valid_inds[0],
                axis=1
            )
            valid_halo_inds_sats = \
                np.delete( self.valid_halo_inds, ind_main_gal_in_valid_inds[0] )

            # Handle when the only valid halo is the main halo
            if dist_to_all_valid_other_gals.size == 0:
                return -2. * np.ones( (self.n_particles,) )

        else:
            raise Exception(
                "ind_main_gal_in_valid_inds too big, is size {}".format(
                    valid_ind_main_gal.size )
            )

        if not scaled:
            return np.min( dist_to_all_valid_other_gals, axis=1 )

        inds_sat = np.argmin( dist_to_all_valid_other_gals, axis=1 )

        # Now scale
        length_scale_sats = self.ahf_halos_length_scale_pkpc[ valid_halo_inds_sats ]

        # Remove the unyt part of things so that we save properly without
        # having to invoke unyt's saving feature
        if isinstance( length_scale_sats, unyt.array.unyt_array ):
            length_scale_sats = length_scale_sats.to_value()

        dist_to_all_valid_other_gals_scaled = dist_to_all_valid_other_gals/length_scale_sats[np.newaxis,:]

        d_other_gal_scaled = dist_to_all_valid_other_gals_scaled[ np.arange( self.n_particles ), inds_sat ]

        return d_other_gal_scaled

    ########################################################################

    def find_host_id( self, radial_cut_fraction=1. ):
        '''Find the host halos our particles are inside of some radial cut of.
        This is the host ID at a given snapshot, and is not the same as the merger tree halo ID.

        Args:
            radial_cut_fraction (float): A particle is in a halo if it's in radial_cut_fraction*length_scale from the center.

        Returns:
            host_halo (np.array of ints): Shape ( n_particles, ).
                The ID of the least massive substructure the particle's part of.
                If it's -1, then the halo ID is the host ID.
                If it's -2, then that particle is not part of any halo, within radial_cut_fraction*length_scale .
        '''

        # Get the halo ID
        halo_id = self.find_halo_id( radial_cut_fraction )

        ahf_host_id =  self.halo_data.get_data( 'hostHalo', self.snum )

        # Handle the case where we have an empty ahf_halos, because there are no halos at that redshift.
        # In this case, the ID will be -2 throughout
        if ahf_host_id.size == 0:
            return halo_id

        # Setup the default array.
        host_id = np.full( halo_id.shape, -2 )

        # Get the host halo ID for valid halos
        valid_halo_id = halo_id != -2
        host_id[valid_halo_id] = ahf_host_id[ halo_id[valid_halo_id] ]

        # Fix the invalid values
        # (which come from not being associated with any halo)
        host_id_fixed = np.ma.fix_invalid( host_id, fill_value=-2 )

        return host_id_fixed.data.astype( int )

    ########################################################################

    def find_halo_id(
        self,
        radial_cut_fraction = 1.,
        type_of_halo_id = 'halo_id',
    ):
        '''Find the smallest halos our particles are inside of some radial cut
        of (we define this as the halo ID). In the case of using MT halo ID,
        we actually find the most massive our particles are inside some
        radial cut of.

        Args:
            radial_cut_fraction (float):
                A particle is in a halo if it's in radial_cut_fraction*length_scale from the center.

            type_of_halo_id (str):
                If 'halo_id' then this is the halo_id at a given snapshot.
                If 'mt_halo_id' then this is the halo_id according to the
                merger tree.

        Returns:
            halo_id (np.array of ints): Shape ( n_particles, ).
                The ID of the least massive substructure the particle's part of.
                In the case of using the 'mt_halo_id', this is the ID of the
                most massive merger tree halo the particle's part of. If it's
                -2, then that particle is not part of any halo, within
                radial_cut_fraction*length_scale .
        '''

        # Choose parameters of the rest of the function based on what type of halo ID we're using
        if type_of_halo_id == 'halo_id':

            # Get the virial masses. It's okay to leave in comoving, since we're just finding the minimum
            m_vir = self.halo_data.get_data( 'Mvir', self.snum )

            # Handle the case where we have an empty ahf_halos, because there are no halos at that redshift.
            # In this case, the halo ID will be -2 throughout
            if m_vir.size == 0:
                halo_id = np.empty( self.n_particles )
                halo_id.fill( -2. )
                return halo_id

            # Functions that change.
            find_containing_halos_fn = self.find_containing_halos
            arg_extremum_fn = np.argmin
            extremum_fn = np.min

        elif type_of_halo_id == 'mt_halo_id':

            # Functions that change.
            find_containing_halos_fn = self.find_mt_containing_halos
            arg_extremum_fn = np.argmax
            extremum_fn = np.max

            # Get the virial masses. It's okay to leave in comoving, since we're just finding the maximum
            m_vir = self.halo_data.data_reader.get_mtree_halo_quantity(
                quantity = 'Mvir',
                indice = self.snum,
                index = self.mtree_halos_index,
                tag = self.halo_file_tag
            )

        else:
            raise Exception( "Unrecognized type_of_halo_id" )

        # Get the cut
        part_of_halo = find_containing_halos_fn(
            radial_cut_fraction = radial_cut_fraction
        )

        # Mask the data
        tiled_m_vir = np.tile( m_vir, ( self.n_particles, 1 ) )
        tiled_m_vir_ma = np.ma.masked_array( tiled_m_vir, mask=np.invert( part_of_halo ), )

        # Take the extremum of the masked data
        if type_of_halo_id == 'halo_id':
            halo_inds = arg_extremum_fn( tiled_m_vir_ma, axis=1 )
            all_halo_ids = self.halo_data.get_data( 'ID', self.snum )
            halo_id = all_halo_ids[halo_inds]
        elif type_of_halo_id == 'mt_halo_id':
            halo_inds = arg_extremum_fn( tiled_m_vir_ma, axis=1 )
            halo_ids = np.array( sorted( self.halo_data.data_reader.mtree_halos.keys() ) )
            halo_id = halo_ids[halo_inds]

        # Account for the fact that the argmin defaults to 0 when there's nothing there
        mask = extremum_fn( tiled_m_vir_ma, axis=1 ).mask
        halo_id = np.ma.filled( np.ma.masked_array(halo_id, mask=mask), fill_value=-2 )

        return halo_id

    ########################################################################

    def find_containing_halos( self, radial_cut_fraction=1. ):
        '''Find which halos our particles are inside of some radial cut of.

        Args:
            radial_cut_fraction (float): A particle is in a halo if it's in radial_cut_fraction*length_scale from the center.

        Returns:
            part_of_halo (np.array of bools): Shape (n_particles, n_halos).
                If index [i, j] is True, then particle i is inside radial_cut_fraction*length_scale of the jth halo.
        '''

        # Get the radial cut
        radial_cut = radial_cut_fraction*self.ahf_halos_length_scale_pkpc[self.valid_halo_inds]

        # Find the halos that our particles are part of (provided they passed the minimum cut)
        part_of_halo_success = self.dist_to_all_valid_halos < radial_cut[np.newaxis,:]

        # Get the full array out
        part_of_halo = np.zeros( (self.n_particles, self.ahf_halos_length_scale_pkpc.size) ).astype( bool )
        part_of_halo[:,self.valid_halo_inds] = part_of_halo_success

        return part_of_halo

    ########################################################################

    def find_mt_containing_halos( self, radial_cut_fraction=1. ):
        '''Find which MergerTrace halos our particles are inside of some
        radial cut of.

        Args:
            radial_cut_fraction (float):
                A particle is in a halo if it's in
                radial_cut_fraction*length_scale from the center.

        Returns:
            part_of_halo (np.array of bools): Shape (n_particles, n_halos).
                If index [i, j] is True, then particle i is inside
                radial_cut_fraction*length_scale of the jth halo, defined
                via the MergerTrace ID.
        '''

        # Load up the merger tree data
        self.halo_data.data_reader.get_mtree_halos(
            self.mtree_halos_index,
            self.halo_file_tag
        )

        part_of_halo = []
        for halo_id in self.halo_data.data_reader.mtree_halos.keys():
            mtree_halo = self.halo_data.data_reader.mtree_halos[ halo_id ]

            # Only try to get the data if we're in the range we actually have
            # the halos for.
            above_minimum_snap = self.snum >= mtree_halo.index.min()

            # Only try to get the data if we have the minimum stellar mass
            if above_minimum_snap:
                halo_value = (
                    mtree_halo[ self.minimum_criteria ][ self.snum ]
                    / self.min_conversion_factor
                )
                has_minimum_value = halo_value >= self.minimum_value
            else:
                # If it's not at the point where it can be traced, it
                # definitely doesn't have the minimum stellar mass.
                has_minimum_value = False

            # Usual case
            if has_minimum_value:

                # Get the halo position
                halo_pos_comov = np.array([
                    mtree_halo['Xc'][ self.snum ],
                    mtree_halo['Yc'][ self.snum ],
                    mtree_halo['Zc'][ self.snum ],
                ])
                halo_pos = halo_pos_comov/( 1. + self.redshift )/self.hubble

                # Make halo_pos 2D for compatibility with cdist
                halo_pos = halo_pos[np.newaxis]

                # Get the distances
                dist = scipy.spatial.distance.cdist( self.particle_positions, halo_pos )

                # Get the relevant length scale
                if self.mt_length_scale == 'r_scale':
                    # Get the scale radius
                    r_vir = mtree_halo[self.halo_length_scale][ self.snum ]
                    length_scale = r_vir/mtree_halo['cAnalytic'][ self.snum ]
                else:
                    length_scale = mtree_halo[self.mt_length_scale][self.snum]
                length_scale_pkpc = length_scale/( 1. + self.redshift )/self.hubble

                # Get the radial distance
                radial_cut = radial_cut_fraction*length_scale_pkpc

                # Find if our particles are part of this halo
                part_of_this_halo = dist < radial_cut

            # Case where there isn't a main halo at that redshift.
            else:
                part_of_this_halo = np.zeros( (self.n_particles, 1) ).astype( bool )

            part_of_halo.append( part_of_this_halo )

        # Format part_of_halo correctly
        part_of_halo = np.array( part_of_halo ).transpose()[0]

        return part_of_halo

    ########################################################################
    # Routines for Calculating Properties of Galaxies
    ########################################################################

    @property
    def mass_inside_galaxy_cut( self ):
        '''
        Returns:
            mass_inside_galaxy_cut (np.ndarray) :
                Mass inside the galaxy_cut*length_scale for *.AHF_halos halos that containing a galaxy.
        '''

        if not hasattr( self, '_mass_inside_galaxy_cut' ):

            self._mass_inside_galaxy_cut = self.summed_quantity_inside_galaxy_valid_halos( self.particle_masses, 0. )

        return self._mass_inside_galaxy_cut

    ########################################################################

    @property
    def mass_inside_all_halos( self ):
        '''
        Returns:
            mass_inside_galaxy_cut (np.ndarray) :
                Mass inside the galaxy_cut*length_scale for all *.AHF_halos halos.
        '''

        if not hasattr( self, '_mass_inside_all_halos' ):
            self._mass_inside_all_halos = np.empty( self.ahf_halos_length_scale_pkpc.shape )
            self._mass_inside_all_halos.fill( np.nan )

            # Case where no halos meet the minimum criteria yet.
            if self.valid_halo_inds.size == 0:
                return self._mass_inside_all_halos

            self._mass_inside_all_halos[self.valid_halo_inds] = self.mass_inside_galaxy_cut

        return self._mass_inside_all_halos

    ########################################################################

    @property
    def cumulative_mass_valid_halos( self ):
        '''Get the cumulative mass, going outwards from the center of each valid *AHF_halos halo.

        Returns:
            cumulative_mass (np.ndarray) : Shape (n_particles, n_halos)
                Index (i,j) is the cumulative mass in halo j at the location of particle i.
        '''

        if not hasattr( self, '_cumulative_mass_valid_halos' ):
            # Sort the mass according to distance
            sorted_inds = np.argsort( self.dist_to_all_valid_halos, axis=0 )
            sorted_mass = self.particle_masses[sorted_inds]

            sorted_cumulative_mass = np.cumsum( sorted_mass, axis=0 )

            # Undo the sorting, now that we have the cumulative mass
            undo_inds = np.argsort( sorted_inds, axis=0 )
            cumulative_mass = sorted_cumulative_mass[undo_inds, np.arange(undo_inds.shape[1])]

            self._cumulative_mass_valid_halos = cumulative_mass

        return self._cumulative_mass_valid_halos

    ########################################################################

    def get_mass_radius( self, mass_fraction ):
        '''Get the radius at which mass_fraction*mass_inside_galaxy_cut is exceeded.

        Args:
            mass_fraction (float) : Mass fraction to consider.

        Returns:
            mass_radius (np.ndarray) :
                The ith index is the distance from the center of halo i to the center of the last particle before
                mass_fraction*mass_inside_galaxy_cut is exceeded.
        '''

        mass_allowed_in_halo = mass_fraction*self.mass_inside_galaxy_cut[np.newaxis,:]
        greater_than_mass_fraction = self.cumulative_mass_valid_halos > mass_allowed_in_halo

        dist_ma = np.ma.masked_array( self.dist_to_all_valid_halos, mask=greater_than_mass_fraction )

        # Fill values in where there's insufficient stellar mass inside that halo, or where we can't resolve the radii.
        # This should be unnecessary if requiring a minimum amount of stellar mass, but it doesn't hurt.
        mass_radius_valid_inds = dist_ma.max( axis=0 )
        mass_radius_valid_inds.fill_value = np.nan
        mass_radius_valid_inds = mass_radius_valid_inds.filled()

        # Now get the mass radius for the full thing.
        mass_radius = np.empty( self.ahf_halos_length_scale_pkpc.shape )
        mass_radius.fill( np.nan )
        mass_radius[self.valid_halo_inds] = mass_radius_valid_inds

        # Convert back to comoving kpc/h
        mass_radius *= ( 1. + self.redshift )*self.hubble

        return mass_radius

    ########################################################################

    def summed_quantity_inside_galaxy_valid_halos( self, particle_quantities, fill_value=0. ):
        '''Get sum( particles_quantities ) for each galaxy (i.e. for particles fulfilling the galaxy cut requirements),
        for halos that are "valid" (i.e. usually meeting some minimum criteria).

        Args:
            particle_quantities (np.ndarray) :
                Quantities to sum.

            fill_value (float) :
                When there are are no particles in a galaxy, what value should be used?

        Returns:
            summed_quantity_inside_galaxy_valid (np.ndarray)
        '''

        def work_fn( particle_quantities_passed, particle_positions_passed=None ):
            '''Private function that actually does what we want.
            '''

            # Case where there are no halos formed yet.
            if self.ahf_halos_length_scale_pkpc.size == 0:
                return np.array( [] )

            # Case where no halos meet the minimum criteria yet.
            if self.valid_halo_inds.size == 0:
                return np.array( [] )

            # If we don't have any particles, just return a filled array.
            if particle_quantities_passed.size == 0:
                return np.array( [ fill_value, ]*self.valid_halo_inds.size )

            if self.low_memory_mode:
                dist_to_all_valid_halos_used = self.dist_to_all_valid_halos_fn( particle_positions_passed )

            else:

                # Make sure we're not trying to use any weird particle positions, i.e. we're using the defaults
                assert particle_positions_passed is None

                dist_to_all_valid_halos_used = self.dist_to_all_valid_halos

            # Make a radial cut.
            valid_radial_cut_pkpc = self.galaxy_cut*self.ahf_halos_length_scale_pkpc[self.valid_halo_inds]
            outside_radial_cut = dist_to_all_valid_halos_used > valid_radial_cut_pkpc[np.newaxis,:]

            # Tile the quantity for proper masking
            quantity_tiled = np.tile( particle_quantities_passed, ( self.valid_halo_inds.size, 1 ) ).transpose()

            quantity_ma = np.ma.masked_array( quantity_tiled, mask=outside_radial_cut )

            # Do the actual sum.
            summed_quantity_inside_galaxy_valid = quantity_ma.sum( axis=0 )

            # Fill in any values where there are no particles in the galaxy.
            summed_quantity_inside_galaxy_valid.fill_value = fill_value
            summed_quantity_inside_galaxy_valid = summed_quantity_inside_galaxy_valid.filled()

            # Make sure that we don't count halos with NaN length scales as containing all particles.
            has_bad_value = np.ma.fix_invalid( valid_radial_cut_pkpc ).mask
            summed_quantity_inside_galaxy_valid = np.where( has_bad_value, np.nan, summed_quantity_inside_galaxy_valid, )

            return summed_quantity_inside_galaxy_valid

        if self.low_memory_mode:

            # Low memory node will not necessarily work for non-zero fill_values, so assert that fill_value == 0
            np.testing.assert_allclose( 0., fill_value, atol=1e-7 )

            # Split the particle quantities into smaller lists
            particle_quantities_chunked = utilities.chunk_list( particle_quantities, self.memory_mode_divisions )
            particle_positions_chunked = utilities.chunk_list( self.particle_positions, self.memory_mode_divisions )

            # Get the result
            summed_quantities_split = [ work_fn( particle_quantities_chunk, particle_positions_chunk ) for \
                                                                    particle_quantities_chunk, particle_positions_chunk in \
                                                                    zip( particle_quantities_chunked, particle_positions_chunked ) ]

            # Sum and return results
            return np.array( summed_quantities_split ).sum( axis=0 )

        else:
            return work_fn( particle_quantities )

    ########################################################################

    def summed_quantity_inside_galaxy( self, particle_quantities, fill_value ):
        '''Get sum( particles_quantities ) for each galaxy (i.e. for particles fulfilling the galaxy cut requirements).
        Args:
            particle_quantities (np.ndarray) :
                Quantities to sum.

            fill_value (float) :
                When there are are no particles in a galaxy, what value should be used?

        Returns:
            summed_quantity_inside_galaxy (np.ndarray)
        '''

        summed_quantity_inside_galaxy = np.empty( self.ahf_halos_length_scale_pkpc.shape )
        summed_quantity_inside_galaxy.fill( np.nan )

        # Case where no halos meet the minimum criteria yet.
        if self.valid_halo_inds.size == 0:
            return summed_quantity_inside_galaxy

        summed_quantity_inside_galaxy[self.valid_halo_inds] = self.summed_quantity_inside_galaxy_valid_halos(
            particle_quantities,
            fill_value,
        )

        return summed_quantity_inside_galaxy

    ########################################################################

    def weighted_summed_quantity_inside_galaxy( self, particle_quantities, particle_weights, fill_value ):
        '''Get sum( particles_quantities ) for each galaxy (i.e. for particles fulfilling the galaxy cut requirements),
        weighted by particle_weights.

        Args:
            particle_quantities (np.ndarray) :
                Quantities to tile.

            particle_weights (np.ndarray) :
                Weights to apply.

            fill_value (float) :
                When there are are no particles in a galaxy, what value should be used?

        Returns:
            summed_quantity_inside_galaxy (np.ndarray)
        '''

        weighted_and_summed = self.summed_quantity_inside_galaxy( particle_quantities*particle_weights, fill_value )
        sum_of_weights = self.summed_quantity_inside_galaxy( particle_weights, fill_value )

        return weighted_and_summed/sum_of_weights






















